/*
 * Classname: ResourceOffering
 * Version: 1.12
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package authentication;

import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class parses a resource offering from an XML node and holds it for other classes to
 * use.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class ResourceOffering {

    private String offeringid;
    private String resourceid;
    private Node offering;
    private String securityMechId;

    /**
     * This constructor stores the input XML node and retrieves a few useful values from it. It
     * also creates the offering identifier by a call to Randomizer.nextID().
     * 
     * @param offering Node The XML node containing the resource offering.
     */
    public ResourceOffering(Node offering) {
        offeringid = Randomizer.nextID();
        this.offering = offering;
        Node idElement = findNode("ResourceID", this.offering);
        resourceid = (idElement != null ? idElement.getTextContent()
                : "urn:liberty:isf:implied-resource");
        Node description = findNode("Description", this.offering);
        if (description != null) {
            Node securityMech = findNode("SecurityMechID", description);
            if (securityMech != null) securityMechId = securityMech.getTextContent();
        }
    }

    /**
     * This method looks for a node with the name withName that is a children of the Node
     * parent. It will return the first matching Node it finds or null if it could not find
     * any.
     * 
     * @param withName String The local name of the node to look for.
     * @param parent Node The parent of the node to look for.
     * @return Node The wanted node or null if it could not be found.
     */
    private Node findNode(String withName, Node parent) {
        NodeList children = parent.getChildNodes();
        Node current;
        for (int i = 0; i < children.getLength(); i++) {
            current = children.item(i);
            if (current.getNodeName().endsWith(withName)) {
                return current;
            } else if (current.hasChildNodes()) {
                return findNode(withName, current);
            }
        }
        return null;
    }

    /**
     * This method appends the resource offering node to the input SOAPElement, after
     * converting the resource offering to a SOAPElement.
     * 
     * @param appendto SOAPElement The element to append this resource offering to.
     * @throws SOAPException If something is wrong with the offering a SOAPException is thrown.
     */
    public void getOffering(SOAPElement appendto) throws SOAPException {
        turnToSoap(offering, appendto);
    }

    /**
     * This method turns the XML node notSoap into a SOAPElement by creating new SOAPElements
     * and giving them the same properties as the XML node. Then it uses recursivity to do the
     * same to notSoap's children.
     */
    private void turnToSoap(Node notSoap, SOAPElement parent) throws SOAPException {
        short nodeType = notSoap.getNodeType();
        if (nodeType != Node.TEXT_NODE && notSoap.getNodeName() != null) {
            SOAPElement soap = parent.addChildElement(notSoap.getNodeName());
            
            if (notSoap.getNodeName().endsWith("Description")) {
                /* Add the credential reference to this element: */
                SOAPElement credentialref = soap.addChildElement("CredentialRef");
                credentialref.addTextNode(offeringid);
            }
            NamedNodeMap attributes = notSoap.getAttributes();
            if (attributes != null) {
                /* Add all the attributes to this element: */
                for (int i = 0; i < attributes.getLength(); i++) {
                    soap.setAttribute(attributes.item(i).getNodeName(), attributes.item(i).getTextContent());
                }
            }

            NodeList children = notSoap.getChildNodes();
            
            /* Add all the children to this element: */
            for (int i = 0; i < children.getLength(); i++) {
                turnToSoap(children.item(i), soap);
            }
        } else if (nodeType == Node.TEXT_NODE) {
            /* Add the text content to this element: */
            String value = notSoap.getTextContent();
            
            if (value != null && value.length() > 0) {
                parent.addTextNode(value);
            }
        }
    }

    /**
     * @return String The unique identifier of this resource offering.
     */
    public String getOfferingId() {
        return offeringid;
    }

    /**
     * @return String The resource identifier contained in this resource offering.
     */
    public String getResourceId() {
        return resourceid;
    }

    /**
     * @return String The message authentication method.
     */
    public String getMessageAuthMethod() {
        /* 
         * Returns the message authentication method that is to be used in communication with the
         * entity represented by this offering. At the time it returns only a predefined method as
         * we only use one. However if several methods needs to be supported it should be changed
         * to return the value of securityMechId after the last ":".
         */
        return "Bearer";
    }
}
