/*
 * Classname: SaslRequest
 * Version: 1.8
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package authentication;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;

import javax.xml.soap.*;

import org.kobjects.isodate.IsoDate;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * This class parses a SASL request from a SOAP message and holds it for other classes to use.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class SaslRequest {

    private static final long MAX_TIME_DEVIATION = 300000;

    private String userid;
    private String messageid;
    private String password;
    private String mechanism;
    private Date timestamp;
    private SOAPMessage message;

    /**
     * The SaslRequest constructor. Starts the parsing and validation of the SOAP message.
     * 
     * @param message SOAPMessage The SOAP message containing the SASL request.
     * @throws Exception If something is wrong with the message.
     */
    public SaslRequest(SOAPMessage message) throws Exception {
        this.message = message;
        parseHead(message.getSOAPHeader());
        parseBody(message.getSOAPBody());
    }

    /**
     * Parses and validates the header fields of the message.
     * 
     * @param header SOAPHeader The header containing the header fields to be checked.
     * @throws Exception If there is something wrong with the header.
     */
    private void parseHead(SOAPHeader header) throws Exception {
        Iterator it = header.extractAllHeaderElements();
        SOAPHeaderElement el;
        while (it.hasNext()) {
            el = (SOAPHeaderElement) it.next();
            if (el.getNodeName().contains("Correlation")) {
                Iterator atts = el.getAllAttributes();
                Name s;
                while (atts.hasNext()) {
                    s = ((Name) atts.next());
                    if (s.getQualifiedName().contains("messageID")) {
                        messageid = el.getAttributeValue(s);
                        messageid.replaceAll(" ", "");
                    } else if (s.getQualifiedName().contains("timestamp")) {
                        timestamp = IsoDate.stringToDate(el.getAttributeValue(s),
                                IsoDate.DATE_TIME);
                        if (Math.abs(timestamp.getTime() - System.currentTimeMillis()) > MAX_TIME_DEVIATION) {
                            throw new Exception();
                        }
                    }
                }
            }
        }

    }

    /**
     * Parses the contents of the SASL request recursively and retrieves the values that are
     * needed later.
     * 
     * @param node Node The node that should be parsed next.
     * @throws Exception If there is something wrong with the request.
     */
    private void parseBody(Node node) throws Exception {
        if (node.getNodeName() != null) {
            if (node.getNodeName().endsWith("SASLRequest") && node.hasAttributes()) {
                NamedNodeMap attributes = node.getAttributes();
                for (int i = 0; i < attributes.getLength(); i++) {
                    if (attributes.item(i).getNodeName().endsWith("mechanism")) {
                        mechanism = attributes.item(i).getTextContent();
                    } else if (attributes.item(i).getNodeName().endsWith("authzID")) {
                        userid = attributes.item(i).getTextContent();
                    }
                }
            }
        }

        if (node.hasChildNodes()) {
            NodeList list = node.getChildNodes();
            for (int i = 0; i < list.getLength(); i++) {
                parseBody(list.item(i));
            }
        } else {
            if (node.getParentNode().getNodeName().endsWith("Data")) {
                String data = node.getNodeValue();
                int startIndex = data.indexOf("%x00") + 4;
                int endIndex = data.indexOf("%x00", startIndex);
                if (userid == null)
                    userid = data.substring(startIndex, endIndex);
                else {
                    String tmpuid = data.substring(startIndex, endIndex);
                    if (!userid.equals(tmpuid)) {
                        /* This request contains inconsistent data */
                        throw new Exception();
                    }
                }
                startIndex = endIndex + 4;
                password = data.substring(startIndex);
            }
        }

    }

    /**
     * @return String The message identifier of this request.
     */
    public String getMessageID() {
        return messageid;
    }

    /**
     * @return String The user identifier in this request.
     */
    public String getUserID() {
        return userid;
    }

    /**
     * @return String The password contained in the request that is supposed to authenticate
     *         the user.
     */
    public String getPassword() {
        return password;
    }

    /**
     * @return Date Returns the timestamp of this request as a Date object so that it is easy
     *         to compare.
     */
    public Date getTimestamp() {
        return timestamp;
    }

    /**
     * @return String Returns the authentication mechanism requested by this request.
     */
    public String getMechanism() {
        return mechanism;
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (AuthenticationService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (AuthenticationService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
