/*
 * Classname: SaslResponse
 * Version: 1.17
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package authentication;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.cert.X509Certificate;

import java.util.Date;

import javax.xml.soap.MessageFactory;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;

import org.kobjects.base64.Base64;
import org.kobjects.isodate.IsoDate;

import org.w3c.dom.NodeList;

/**
 * This class creates the SASL response that is to be returned to the client. This includes
 * adding the ResourceOffering for the discovery service and creating the assertion that allows
 * the user to access this service.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class SaslResponse {

    public static final String STATUS_OK = "as:OK";
    public static final String STATUS_ABORT = "as:abort";
    public static final String STATUS_CONTINUE = "as:continue";

    private static final String disco = "urn:liberty:disco:2003-08";
    private static final String saml = "urn:oasis:names:tc:SAML:1.0:assertion";
    private static final String ds = "http://www.w3.org/2000/09/xmldsig#";
    private static final String sa = "urn:liberty:sa:2004-04";
    private static final String sb = "urn:liberty:sb:2003-08";

    private static final int TIME_TO_LIVE = 900000;
    private static final String AUTH_METHOD_PASSWORD = "urn:oasis:names:tc:SAML:1.0:am:password";
    private static final String keystoreFile = ".asstore";

    private ResourceOffering offering;
    private String inResponseTo;
    private SOAPMessage message;
    private SOAPEnvelope env;
    private String statusCode;
    private X509Certificate cert;
    private PrivateKey privateKey;
    private String issuerid;
    private String serverMech;

    /**
     * The SaslRequest constructor. Initialises the member variables and starts the creation of
     * the header and body elements.
     * 
     * @param inResponseTo String The message identifier of the request this SASL response
     *            should respond to.
     * @param statusCode String The status code of the response.
     * @param serverMech String The SASL mechanism used to authenticate the user.
     * @param offering ResourceOffering The resource offering for the discovery service.
     */
    public SaslResponse(String inResponseTo, String statusCode, String serverMech,
            ResourceOffering offering) {
        this.offering = offering;
        this.inResponseTo = inResponseTo;
        this.statusCode = statusCode;
        this.serverMech = serverMech;
        try {
            /* Create the message that should be sent back to the client: */
            message = MessageFactory.newInstance().createMessage();
            message.setProperty(SOAPMessage.CHARACTER_SET_ENCODING, "UTF-8");
            message.setProperty(SOAPMessage.WRITE_XML_DECLARATION, "true");
            env = message.getSOAPPart().getEnvelope();

            /* Read the certificate and private key from the keystore: */
            KeyStore keystore = KeyStore.getInstance("jks");
            keystore.load(new FileInputStream(keystoreFile), new char[] { '6', '5', '4', '3',
                    '2', '1' });
            cert = (X509Certificate) keystore.getCertificate("idpas");
            issuerid = cert.getIssuerDN().getName();
            privateKey = (PrivateKey) keystore.getKey("idpas", new char[] { '1', '2', '3',
                    '4', '5', '6' });

            /* Fill in the contents of the response: */
            createHeaders();
            createBody();
        } catch (Exception e) {
            /* Something wrong happened. */
            setStatusFailed();
        }
    }

    /**
     * If the response encounters exceptions it will not be able to return a proper response.
     * In these cases a response with the status code abort is returned instead. This method
     * replaces the SaslResponse with a new SaslResponse with the status code abort.
     */
    private void setStatusFailed() {
        this.statusCode = STATUS_ABORT;
        try {
            env = message.getSOAPPart().getEnvelope();
            createHeaders();
            createBody();
        } catch (Exception e) {
            if (AuthenticationService.DEBUG) Log.write(e);
        }
    }

    /**
     * This method creates the header by adding all the required fields such as the correlation
     * and timestamp.
     */
    private void createHeaders() throws SOAPException {
        SOAPHeader header = message.getSOAPHeader();
        SOAPHeaderElement correlation = header.addHeaderElement(env.createName("Correlation",
                "sb", sb));
        correlation.addAttribute(env.createName("messageID", "sb", sb), Randomizer.nextID());
        correlation.addAttribute(env.createName("refToMessageID", "sb", sb), inResponseTo);
        correlation.addAttribute(env.createName("timestamp", "sb", sb), IsoDate.dateToString(
                new Date(System.currentTimeMillis()), IsoDate.DATE_TIME));
        correlation.setMustUnderstand(true);

    }

    /**
     * When creating the body the method adds the status element and an element which holds the
     * content of the response. If the status is OK, then methods to add the offering and token
     * for the discovery service are called.
     */
    private void createBody() throws SOAPException, GeneralSecurityException {
        SOAPBody body = message.getSOAPBody();
        NodeList children = body.getChildNodes();
        if (children != null && children.getLength() <= 0) {
            body.removeContents();
        }
        /* SASLResponse element: */
        SOAPElement saslResponse = body.addChildElement(env.createName("SASLResponse", "sa",
                sa));
        saslResponse.addAttribute(env.createName("serverMechanism", "sa", sa), serverMech);
        /* SASLResponse subelement, Status element: */
        SOAPElement status = saslResponse.addChildElement(env.createName("Status", "sa", sa));
        status.addTextNode(statusCode);
        if (statusCode.equals(STATUS_OK)) {
            offering.getOffering(saslResponse);
            createToken(saslResponse);
        }
    }

    /**
     * This method creates an authentication assertion for the resource offering to the
     * discovery service.
     */
    private void createToken(SOAPElement saslResponse) throws SOAPException,
            GeneralSecurityException {
        if ("Bearer".equals(offering.getMessageAuthMethod())) {

            /* First create all the node names: */

            /* SASLResponse subelement, Credentials element: */
            Name credentialsName = env.createName("Credentials", "disco", disco);
            /* Credentials subelement, Assertion element: */
            Name assertionName = env.createName("Assertion", "saml", saml);
            /* Assertion attributes: */
            Name majorVersionName = env.createName("MajorVersion", "saml", saml);
            Name minorVersionName = env.createName("MinorVersion", "saml", saml);
            Name assertionIdName = env.createName("AssertionID", "saml", saml);
            Name issuerName = env.createName("Issuer", "saml", saml);
            Name issueInstantName = env.createName("IssueInstant", "saml", saml);
            /* Assertion subelement, Conditions element: */
            Name conditionsName = env.createName("Conditions", "saml", saml);
            /* Conditions attributes: */
            Name notBeforeName = env.createName("NotBefore", "saml", saml);
            Name notOnOrAfterName = env.createName("NotOnOrAfter", "saml", saml);
            /* Conditions subelement, AudienceRestrictionCondition element: */
            Name audienceRestrictionName = env.createName("AudienceRestrictionCondition",
                    "saml", saml);
            /* AudienceRestrictionCondition subelement, Audience element: */
            Name audienceName = env.createName("Audience", "saml", saml);
            /* Assertion subelement, AuthenticationStatement element: */
            Name authenticationStatementName = env.createName("AuthenticationStatement",
                    "saml", saml);
            /* AuthenticationStatement attributes: */
            Name authenticationMethodName = env.createName("AuthenticationMethod", "saml",
                    saml);
            Name authenticationInstantName = env.createName("AuthenticationInstant", "saml",
                    saml);

            SOAPElement credentials = saslResponse.addChildElement(credentialsName);
            SOAPElement assertion;
            SOAPElement conditions;
            SOAPElement audienceRestriction;
            SOAPElement audience;
            SOAPElement authenticationStatement;
            long now = System.currentTimeMillis();

            /* Then create the token itself: */

            assertion = credentials.addChildElement(assertionName);
            assertion.addAttribute(majorVersionName, "1");
            assertion.addAttribute(minorVersionName, "1");
            assertion.addAttribute(assertionIdName, offering.getOfferingId());
            assertion.addAttribute(issuerName, issuerid);
            assertion.addAttribute(issueInstantName, IsoDate.dateToString(new Date(now),
                    IsoDate.DATE_TIME));

            conditions = assertion.addChildElement(conditionsName);
            conditions.addAttribute(notBeforeName, IsoDate.dateToString(new Date(now),
                    IsoDate.DATE_TIME));
            conditions.addAttribute(notOnOrAfterName, IsoDate.dateToString(new Date(now
                    + TIME_TO_LIVE), IsoDate.DATE_TIME));

            audienceRestriction = conditions.addChildElement(audienceRestrictionName);

            audience = audienceRestriction.addChildElement(audienceName);
            audience.addTextNode(offering.getResourceId());

            authenticationStatement = assertion.addChildElement(authenticationStatementName);
            authenticationStatement.addAttribute(authenticationMethodName,
                    AUTH_METHOD_PASSWORD);
            authenticationStatement.addAttribute(authenticationInstantName, IsoDate
                    .dateToString(new Date(now), IsoDate.DATE_TIME));

            createSignature(assertion);
        }
    }

    /**
     * This method signs an authentication assertion without the signature element and adds the
     * signature to the assertion.
     */
    private void createSignature(SOAPElement assertion) throws SOAPException,
            GeneralSecurityException {
        byte[] text = assertion.toString().getBytes();
        MessageDigest mac = MessageDigest.getInstance("SHA-1");
        mac.update(text, 0, text.length);
        byte[] digested = mac.digest();
        String digestString = Base64.encode(digested);

        Signature signer = Signature.getInstance("SHA1withRSA");
        signer.initSign(privateKey);
        signer.update(text);
        byte[] signed = signer.sign();
        String signatureString = Base64.encode(signed);

        /* Assertion subelement, Signature element: */
        SOAPElement signature = assertion.addChildElement(env
                .createName("Signature", "ds", ds));
        /* Signature subelement, SignedInfo element: */
        SOAPElement signedInfo = signature.addChildElement(env.createName("SignedInfo", "ds",
                ds));
        /* SignedInfo subelement, CanonicalizationMethod element: */
        SOAPElement canonMethod = signedInfo.addChildElement(env.createName(
                "CanonicalizationMethod", "ds", ds));
        /* CanonicalizationMethod attributes: */
        canonMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/TR/2001/REC-xml-c14n-20010315#");
        /* SignedInfo subelement, SignatureMethod element: */
        SOAPElement signatureMethod = signedInfo.addChildElement(env.createName(
                "SignatureMethod", "ds", ds));
        /* SignatureMethod attributes: */
        signatureMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/2000/09/xmldsig#rsa-sha1");
        /* SignatureMethod subelement, HMACOutputLength element: */
        SOAPElement hmacOutputLength = signatureMethod.addChildElement(env.createName(
                "HMACOutputLength", "ds", ds));
        hmacOutputLength.addTextNode(mac.getDigestLength() + "");
        /* SignedInfo subelement, Reference element: */
        SOAPElement reference = signedInfo.addChildElement(env.createName("Reference", "ds",
                ds));
        /* Reference subelement, DigestMethod element: */
        SOAPElement digestMethod = reference.addChildElement(env.createName("DigestMethod",
                "ds", ds));
        /* DigestMethod attributes: */
        digestMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/2000/09/xmldsig#sha1");
        /* Reference subelement, DigestValue element: */
        SOAPElement digestValue = reference.addChildElement(env.createName("DigestValue",
                "ds", ds));
        digestValue.addTextNode(digestString);
        /* Signature subelement, SignatureValue element: */
        SOAPElement signatureValue = signature.addChildElement(env.createName(
                "SignatureValue", "ds", ds));
        signatureValue.addTextNode(signatureString);
        /* Signature subelement, KeyInfo element: */
        SOAPElement keyInfo = signature.addChildElement(env.createName("KeyInfo", "ds", ds));
        /* KeyInfo subelement, X509Data element: */
        SOAPElement x509data = keyInfo.addChildElement(env.createName("X509Data", "ds", ds));
        /* X509Data subelement, X509Certificate element: */
        SOAPElement x509certificate = x509data.addChildElement(env.createName(
                "X509Certificate", "ds", ds));
        x509certificate.addTextNode(Base64.encode(cert.getEncoded()));

    }

    /**
     * @return SOAPMessage The SOAP message containing this response.
     */
    public SOAPMessage getMessage() {
        return message;
    }

    /**
     * Returns a String representation of the SOAP message containing this response.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (AuthenticationService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (AuthenticationService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
