/*
 * Classname: DiscoQuery
 * Version: 1.18
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package discovery;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.Signature;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Iterator;

import javax.xml.soap.Name;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;

import org.kobjects.base64.Base64;
import org.kobjects.isodate.IsoDate;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The DiscoQuery is used to parse the request and save the information that we need to verify
 * the validity of the message and to find the information needed in the response.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class DiscoQuery {

    private static final long MAX_TIME_DEVIATION = 300000;
    private static final long SESSION_TIME_TO_LIVE = 600000;

    private Date sessiontimeout;
    private Date timestamp;
    private Node authenticationAssertion;
    private Node authenticationAssertionCopy;
    private String messageid;
    private String resourceid;
    private String userid;
    private String reqServiceType;
    private String authenticationIssuer;
    private SOAPMessage message;

    /**
     * The constructor tries to parse the body and the header of the SOAP message.
     * 
     * @param message SOAPMessage The SOAPMessage containing the request
     * @throws Exception Thrown to show that the request parsing was unsuccessful
     */
    public DiscoQuery(SOAPMessage message) throws Exception {
        this.message = message;
        parseBody(message.getSOAPBody());
        parseHead(message.getSOAPHeader());
    }

    /**
     * This method runs through all the nodes in the request body to find the nodes we need,
     * which are the service type and the resourceID.
     */
    private void parseBody(Node node) {
        if (node.hasChildNodes()) {
            NodeList list = node.getChildNodes();
            Node child;
            for (int i = 0; i < list.getLength(); i++) {
                child = list.item(i);
                if (child != null) {
                    if (child.getNodeName().endsWith("ServiceType")) {
                        reqServiceType = child.getFirstChild().getNodeValue();
                    } else if (child.getNodeName().endsWith("ResourceID")) {
                        resourceid = child.getFirstChild().getNodeValue();
                    } else {
                        parseBody(child);
                    }
                }
            }
        }
    }

    /**
     * This method runs through all the nodes in the request body to find the nodes we need,
     * which are the timestamp and the message token. The security token is checked in a
     * separate method.
     */
    private void parseHead(SOAPHeader node) throws Exception {
        if (node instanceof SOAPHeader) {
            SOAPHeader header = (SOAPHeader) node;
            Iterator it = header.extractAllHeaderElements();
            SOAPHeaderElement el;
            while (it.hasNext()) {
                el = (SOAPHeaderElement) it.next();
                if (el.getNodeName().contains("Correlation")) {
                    Iterator atts = el.getAllAttributes();
                    Name s;
                    while (atts.hasNext()) {
                        s = ((Name) atts.next());
                        if (s.getQualifiedName().contains("messageID"))
                            messageid = el.getAttributeValue(s);
                        else if (s.getQualifiedName().contains("timestamp")) {
                            timestamp = IsoDate.stringToDate(el.getAttributeValue(s),
                                    IsoDate.DATE_TIME);
                            if (Math.abs(timestamp.getTime() - System.currentTimeMillis()) > MAX_TIME_DEVIATION) {
                                /*
                                 * This could be replay, but we should store previous
                                 * messageIDs to check that
                                 */
                                throw new Exception("Timestamp deviation");
                            }
                        }
                    }
                } else if (el.getNodeName().contains("Security")) {
                    processSecToken(el, 0);
                }
            }
        }

    }

    /**
     * This method processes the security token to find if the assertion from the
     * authentication service is correct. It uses a separate method to validate the signature
     * on the assertion.
     */
    private void processSecToken(Node token, int parsingDepth) throws Exception {
        int nextDebth = parsingDepth + 1;
        if (token.hasChildNodes()) {
            NodeList list = token.getChildNodes();
            Node child;
            for (int i = 0; i < list.getLength(); i++) {
                child = list.item(i);
                if (child.getNodeName() == null) {
                    /* Do nothing */
                } else if (child.getNodeName().endsWith("Assertion")) {
                    authenticationAssertion = child.cloneNode(true);
                    authenticationAssertionCopy = child.cloneNode(true);
                    NamedNodeMap atts = child.getAttributes();
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("Issuer")) {
                            authenticationIssuer = atts.item(j).getTextContent();
                        }
                    }
                } else if (child.getNodeName().endsWith("Conditions")) {
                    NamedNodeMap atts = child.getAttributes();
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("NotBefore")) {
                            Date notBefore = IsoDate.stringToDate(atts.item(j)
                                    .getTextContent(), IsoDate.DATE_TIME);
                            if (notBefore.after(new Date(System.currentTimeMillis()))) {
                                throw new Exception();
                            }
                        } else if (atts.item(j).getNodeName().endsWith("NotOnOrAfter")) {
                            Date notOnOrAfter = IsoDate.stringToDate(atts.item(j)
                                    .getTextContent(), IsoDate.DATE_TIME);
                            if (notOnOrAfter.before(new Date(System.currentTimeMillis()))) {
                                throw new Exception();
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("Audience")) {
                    if (!child.getFirstChild().getNodeValue().equals(resourceid)) {
                        throw new Exception();
                    }
                } else if (child.getNodeName().endsWith("AuthenticationStatement")) {
                    NamedNodeMap atts = child.getAttributes();
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("AuthenticationInstant")) {
                            sessiontimeout = new Date(IsoDate.stringToDate(
                                    atts.item(j).getTextContent(), IsoDate.DATE_TIME)
                                    .getTime()
                                    + SESSION_TIME_TO_LIVE);
                            if (sessionTimedOut()) {
                                throw new Exception(sessiontimeout.toString());
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("Signature")) {
                    validateSignature(child);
                }
                processSecToken(child, nextDebth);
            }
        }
        if ((sessiontimeout == null || authenticationIssuer == null) && parsingDepth == 0) {
            throw new Exception();
        }
    }

    /**
     * This method removes extra spaces, line breaks and the signature element to make the
     * assertion similar at all times when using it for signing and signature verification.
     */
    private String prepareAssertion(String assertion) {
        StringBuffer buffer = new StringBuffer(assertion);

        /* Removes all multiple spaces: */
        int doubleSpace = buffer.indexOf("  ");
        int somethingElse;
        while (doubleSpace > -1) {
            somethingElse = buffer.indexOf("<", doubleSpace);
            buffer.replace(doubleSpace, somethingElse, "");
            doubleSpace = buffer.indexOf("  ", doubleSpace);
        }

        /* Removes all \n: */
        int linebreak = buffer.indexOf("\n");
        while (linebreak > -1) {
            somethingElse = buffer.indexOf("<", linebreak);
            buffer.replace(linebreak, somethingElse, "");
            linebreak = buffer.indexOf("\n", linebreak);
        }

        /* Removes all \r: */
        linebreak = buffer.indexOf("\r");
        while (linebreak > -1) {
            somethingElse = buffer.indexOf("<", linebreak);
            buffer.replace(linebreak, somethingElse, "");
            linebreak = buffer.indexOf("\r", linebreak);
        }

        assertion = buffer.toString();

        /* Removes the signature element */
        int signatureStartIndex = assertion.indexOf("Signature");
        int tagStartIndex = assertion.indexOf("<", signatureStartIndex - 10);
        int lastTagStartIndex = 0;
        while (tagStartIndex < signatureStartIndex) {
            lastTagStartIndex = tagStartIndex;
            tagStartIndex = assertion.indexOf("<", lastTagStartIndex + 1);
        }
        tagStartIndex = lastTagStartIndex;
        int signatureEndIndex = assertion.indexOf("Signature>", signatureStartIndex + 1)
                + "Signature>".length();
        String result = assertion.substring(0, tagStartIndex)
                + assertion.substring(signatureEndIndex, assertion.length());

        return result;
    }

    /**
     * This method validates the signature and throws an Exception if there is something wrong,
     * to indicate that an invalid assertion and that we should not proceed with creating a
     * response.
     */
    private void validateSignature(Node signaturenode) throws Exception {
        String clearText = prepareAssertion(authenticationAssertionCopy.toString());
        NodeList children = signaturenode.getChildNodes();
        Node child;
        X509Certificate cert = null;
        String signatureString = null;
        for (int i = 0; i < children.getLength(); i++) {
            child = children.item(i);
            if (child.getNodeName().endsWith("SignatureValue")) {
                /* This node contains the actual signature */
                signatureString = child.getFirstChild().getNodeValue();
            } else if (child.getNodeName().endsWith("KeyInfo")) {
                /* This node contains information about the key used to create the signature */
                Node x509data = child.getFirstChild();
                Node x509certificate = x509data.getFirstChild();
                String certAsString = x509certificate.getFirstChild().getNodeValue();
                CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
                ByteArrayInputStream certInput = new ByteArrayInputStream(Base64.decode(certAsString));
                cert = (X509Certificate) certFactory.generateCertificate(certInput);

                try {
                    certInput.close();
                } catch (IOException e) {
                }
                String certificateissuer = cert.getIssuerDN().getName();

                if (!authenticationIssuer.equals(certificateissuer)) {
                    GeneralSecurityException e = new GeneralSecurityException();
                    if (DiscoveryService.DEBUG) Log.write(e);
                    throw e;
                }
            }
        }

        Signature signer = Signature.getInstance("SHA1withRSA");
        signer.initVerify(cert);
        signer.update(clearText.getBytes());
        boolean ok = signer.verify(Base64.decode(signatureString));

        if (!ok) {
            GeneralSecurityException e = new GeneralSecurityException();
            throw e;
        }
    }

    /**
     * @return String Returns the userID.
     */
    public String getUserID() {
        if (userid == null){
            userid = resourceid.substring(resourceid.lastIndexOf("/") + 1);
        }
        return userid;
    }

    /**
     * @return String Returns the resourceID.
     */
    public String getResourceID() {
        return resourceid;
    }

    /**
     * @return String Returns the messageID
     */
    public String getMessageID() {
        return messageid;
    }

    /**
     * @return Date Returns the timestamp of the message.
     */
    public Date getTimestamp() {
        return timestamp;
    }

    /**
     * @return String Returns the service type identified in the request.
     */
    public String getReqServiceType() {
        return reqServiceType;
    }

    /**
     * @return Node Returns the assertion from the authentication service which was enclosed in
     *         the request.
     */
    public Node getAuthenticationAssertion() {
        return authenticationAssertion.cloneNode(true);
    }

    /**
     * @return boolean Returns true if the session has timed out and false if not.
     */
    public boolean sessionTimedOut() {
        return sessiontimeout.before(new Date(System.currentTimeMillis()));
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (DiscoveryService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (DiscoveryService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
