/*
 * Classname: DiscoResponse
 * Version: 1.16
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package discovery;

import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.KeyStore;
import java.security.MessageDigest;
import java.security.PrivateKey;
import java.security.Signature;
import java.security.cert.X509Certificate;
import java.util.Date;

import javax.xml.soap.MessageFactory;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;

import org.kobjects.base64.Base64;
import org.kobjects.isodate.IsoDate;

/**
 * The class DiscoResponse is used to create the reponse message. It needs to create the header
 * and the body. The body contains the offerings or an status code indicating error. This also
 * creates the tokens and a signature of the assertion.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class DiscoResponse {

    public static final String STATUS_OK = "disco:OK";
    public static final String STATUS_FAILED = "disco:Failed";

    private static final String disco = "urn:liberty:disco:2003-08";
    private static final String saml = "urn:oasis:names:tc:SAML:1.0:assertion";
    private static final String ds = "http://www.w3.org/2000/09/xmldsig#";

    private static final int TIME_TO_LIVE = 900000;

    private static final String keystoreFile = ".dsstore";

    private DiscoQuery query;
    private PrivateKey privateKey;
    private ResourceOffering[] offerings;
    private String statusCode;
    private String issuerid;
    private SOAPMessage message;
    private SOAPEnvelope env;
    private X509Certificate cert;

    /**
     * The constructor creates the message and initialises variables that are used in message
     * creation. It calls two other methods to create the header and body of the message.
     * 
     * @param query DiscoQuery The request message related to this response
     * @param offerings ResourceOffering[] The offerings that was found for this user
     * @param statusCode String The status code which indicates the type of message to be sent
     */
    public DiscoResponse(DiscoQuery query, ResourceOffering[] offerings, String statusCode) {
        this.offerings = offerings;
        this.query = query;
        this.statusCode = statusCode;
        try {
            message = MessageFactory.newInstance().createMessage();
            message.setProperty(SOAPMessage.CHARACTER_SET_ENCODING, "UTF-8");
            message.setProperty(SOAPMessage.WRITE_XML_DECLARATION, "true");

            KeyStore keystore = KeyStore.getInstance("jks");
            keystore.load(new FileInputStream(keystoreFile), new char[] { '6', '5', '4', '3',
                    '2', '1' });
            cert = (X509Certificate) keystore.getCertificate("idpds");
            issuerid = cert.getIssuerDN().getName();
            privateKey = (PrivateKey) keystore.getKey("idpds", new char[] { '1', '2', '3', '4', '5', '6' });

            env = message.getSOAPPart().getEnvelope();
            createHeaders();
            createBody();
        } catch (Exception e) {
            setStatusFailed();
        }
    }

    /**
     * Used to set the status code of this message to failed if something went wrong during the
     * message creation. This also removes anything added to the message and creates an error
     * message.
     */
    private void setStatusFailed() {
        this.statusCode = STATUS_FAILED;
        try {
            env = message.getSOAPPart().getEnvelope();
            createHeaders();
            try {
                createBody();
            } catch (Exception e2) {
                if (DiscoveryService.DEBUG) Log.write(e2);
            }
        } catch (SOAPException e1) {
            if (DiscoveryService.DEBUG) Log.write(e1);
        }
    }

    /**
     * This method creates the header by adding all the required fields such as the correlation
     * and timestamp.
     */
    private void createHeaders() throws SOAPException {
        SOAPHeader header = message.getSOAPHeader();
        SOAPHeaderElement correlation = header.addHeaderElement(env.createName("Correlation",
                "sb", "urn:liberty:sb:2003-08"));
        correlation.addAttribute(env.createName("messageID", "sb", "urn:liberty:sb:2003-08"),
                Randomizer.nextID());
        if (query != null) {
            String refTo = query.getMessageID();
            if (refTo != null) {
                correlation.addAttribute(env.createName("refToMessageID", "sb",
                        "urn:liberty:sb:2003-08"), refTo);
            }
        }
        correlation.addAttribute(env.createName("timestamp", "sb", "urn:liberty:sb:2003-08"),
                IsoDate.dateToString(new Date(System.currentTimeMillis()), IsoDate.DATE_TIME));
        correlation.setMustUnderstand(true);
    }

    /**
     * When creating the body the method adds the status element and an element which holds the
     * content of the response. If the status is OK, then methods to add the offerings and
     * tokens are called.
     */
    private void createBody() throws SOAPException, GeneralSecurityException {
        SOAPBody body = message.getSOAPBody();

        /* QueryResponse element: */
        SOAPElement queryResponse = body.addChildElement(env.createName("QueryResponse", "disco", disco));
        
        /* QueryResponse subelement, Status element: */
        SOAPElement status = queryResponse.addChildElement(env.createName("Status", "disco", disco));
        if (statusCode.equals(STATUS_OK) && offerings != null) {
            status.addTextNode(statusCode);
            addOfferings(queryResponse);
            createTokens(queryResponse);
        } else {
            status.addTextNode(STATUS_FAILED);
        }
    }

    /**
     * This appends the offerings to the response element by calling a method in
     * ResourceOffering which adds offerings (stored in a Node) to a SOAPElement.
     */
    private void addOfferings(SOAPElement queryResponse) throws SOAPException {
        for (int i = 0; i < offerings.length; i++) {
            offerings[i].getOffering(queryResponse);
        }
    }

    /**
     * This method creates an authorisation assertion for each resource offering returned.
     */
    private void createTokens(SOAPElement queryResponse) throws SOAPException,
            GeneralSecurityException {

        /* First create all the node names: */

        /* SASLResponse subelement, Credentials element: */
        Name credentialsName = env.createName("Credentials", "disco", disco);
        /* Credentials subelement, Assertion element: */
        Name assertionName = env.createName("Assertion", "saml", saml);
        /* Assertion attributes: */
        Name majorVersionName = env.createName("MajorVersion", "saml", saml);
        Name minorVersionName = env.createName("MinorVersion", "saml", saml);
        Name assertionIdName = env.createName("AssertionID", "saml", saml);
        Name issuerName = env.createName("Issuer", "saml", saml);
        Name issueInstantName = env.createName("IssueInstant", "saml", saml);
        /* Assertion subelement, Conditions element: */
        Name conditionsName = env.createName("Conditions", "saml", saml);
        /* Conditions attributes: */
        Name notBeforeName = env.createName("NotBefore", "saml", saml);
        Name notOnOrAfterName = env.createName("NotOnOrAfter", "saml", saml);
        /* Conditions subelement, AudienceRestrictionCondition element: */
        Name audienceRestrictionName = env.createName("AudienceRestrictionCondition", "saml",
                saml);
        /* AudienceRestrictionCondition subelement, Audience element: */
        Name audienceName = env.createName("Audience", "saml", saml);
        /* Assertion subelement, AuthorizationDecisionStatement element: */
        Name authorizationDecisionStatementName = env.createName(
                "AuthorizationDecisionStatement", "saml", saml);
        /* AuthorizationDecisionStatement attributes: */
        Name resourceName = env.createName("Resource", "saml", saml);
        Name decisionName = env.createName("Decision", "saml", saml);
        /* AuthorizationDecisionStatement subelement, Action element: */
        Name actionName = env.createName("Action", "saml", saml);
        /* Action attributes: */
        Name namespaceName = env.createName("Namespace", "saml", saml);
        /* AuthorizationDecisionStatement subelement, Evidence element: */
        Name evidenceName = env.createName("Evidence", "saml", saml);

        SOAPElement credentials = queryResponse.addChildElement(credentialsName);
        SOAPElement assertion;
        SOAPElement conditions;
        SOAPElement audienceRestriction;
        SOAPElement audience;
        SOAPElement authorizationDecisionStatement;
        SOAPElement action;
        SOAPElement evidence;

        /* Then we create the nodes and add values: */

        for (int i = 0; i < offerings.length; i++) {
            if ("Bearer".equals(offerings[i].getMessageAuthMethod())) {
                long now = System.currentTimeMillis();
                assertion = credentials.addChildElement(assertionName);
                assertion.addAttribute(majorVersionName, "1");
                assertion.addAttribute(minorVersionName, "1");
                assertion.addAttribute(assertionIdName, offerings[i].getOfferingId());
                assertion.addAttribute(issuerName, issuerid);
                assertion.addAttribute(issueInstantName, IsoDate.dateToString(new Date(now),
                        IsoDate.DATE_TIME));

                conditions = assertion.addChildElement(conditionsName);
                conditions.addAttribute(notBeforeName, IsoDate.dateToString(new Date(now), IsoDate.DATE_TIME));
                conditions.addAttribute(notOnOrAfterName, IsoDate.dateToString(new Date(now
                        + TIME_TO_LIVE), IsoDate.DATE_TIME));

                audienceRestriction = conditions.addChildElement(audienceRestrictionName);

                audience = audienceRestriction.addChildElement(audienceName);
                audience.addTextNode(offerings[i].getResourceId());

                authorizationDecisionStatement = assertion.addChildElement(authorizationDecisionStatementName);
                authorizationDecisionStatement.addAttribute(resourceName, offerings[i].getResourceId());
                authorizationDecisionStatement.addAttribute(decisionName, "Permit");

                action = authorizationDecisionStatement.addChildElement(actionName);
                action.addAttribute(namespaceName, "urn:oasis:names:tc:SAML:1.0:action:rwedc");
                action.addTextNode("Read");

                evidence = authorizationDecisionStatement.addChildElement(evidenceName);
                /* Add the authentication assertion as evidence: */
                evidence.appendChild(query.getAuthenticationAssertion());

                createSignature(assertion);
            }
        }
    }

    /**
     * This method removes line breaks to make the assertion similar at all times when using it
     * for signing and signature verification.
     */
    private String prepareAssertion(String assertion) {
        StringBuffer buffer = new StringBuffer(assertion);

        /* Removes all \n: */
        int linebreak = buffer.indexOf("\n");
        while (linebreak > -1) {
            buffer.replace(linebreak, linebreak + 1, "");
            linebreak = buffer.indexOf("\n", linebreak);
        }

        /* Removes all \r: */
        linebreak = buffer.indexOf("\r");
        while (linebreak > -1) {
            buffer.replace(linebreak, linebreak + 1, "");
            linebreak = buffer.indexOf("\r", linebreak);
        }

        return buffer.toString();
    }

    /**
     * This method signs an authorisation assertion without the signature element and adds the
     * signature to the assertion.
     */
    private void createSignature(SOAPElement assertion) throws SOAPException, GeneralSecurityException {
        byte[] text = prepareAssertion(assertion.toString()).getBytes();

        /* This creates an assertion digest: */
        MessageDigest mac = MessageDigest.getInstance("SHA-1");
        mac.update(text, 0, text.length);
        byte[] digested = mac.digest();
        String digestString = Base64.encode(digested);

        /* This creates an assertion signature: */
        Signature signer = Signature.getInstance("SHA1withRSA");
        signer.initSign(privateKey);
        signer.update(text);
        byte[] signed = signer.sign();
        String signatureString = Base64.encode(signed);

        /* Assertion subelement, Signature element: */
        SOAPElement signature = assertion.addChildElement(env.createName("Signature", "ds", ds));
        /* Signature subelement, SignedInfo element: */
        SOAPElement signedInfo = signature.addChildElement(env.createName("SignedInfo", "ds", ds));
        /* SignedInfo subelement, CanonicalizationMethod element: */
        SOAPElement canonMethod = signedInfo.addChildElement(env.createName(
                "CanonicalizationMethod", "ds", ds));
        /* CanonicalizationMethod attributes: */
        canonMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/TR/2001/REC-xml-c14n-20010315#");
        /* SignedInfo subelement, SignatureMethod element: */
        SOAPElement signatureMethod = signedInfo.addChildElement(env.createName(
                "SignatureMethod", "ds", ds));
        /* SignatureMethod attributes: */
        signatureMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/2000/09/xmldsig#rsa-sha1");
        /* SignatureMethod subelement, HMACOutputLength element: */
        SOAPElement hmacOutputLength = signatureMethod.addChildElement(env.createName(
                "HMACOutputLength", "ds", ds));
        hmacOutputLength.addTextNode(mac.getDigestLength() + "");
        /* SignedInfo subelement, Reference element: */
        SOAPElement reference = signedInfo.addChildElement(env.createName("Reference", "ds", ds));
        /* Reference subelement, DigestMethod element: */
        SOAPElement digestMethod = reference.addChildElement(env.createName("DigestMethod", "ds", ds));
        /* DigestMethod attributes: */
        digestMethod.addAttribute(env.createName("Algorithm", "ds", ds),
                "http://www.w3.org/2000/09/xmldsig#sha1");
        /* Reference subelement, DigestValue element: */
        SOAPElement digestValue = reference.addChildElement(env.createName("DigestValue", "ds", ds));
        digestValue.addTextNode(digestString);
        /* Signature subelement, SignatureValue element: */
        SOAPElement signatureValue = signature.addChildElement(env.createName(
                "SignatureValue", "ds", ds));
        signatureValue.addTextNode(signatureString);
        /* Signature subelement, KeyInfo element: */
        SOAPElement keyInfo = signature.addChildElement(env.createName("KeyInfo", "ds", ds));
        /* KeyInfo subelement, X509Data element: */
        SOAPElement x509data = keyInfo.addChildElement(env.createName("X509Data", "ds", ds));
        /* X509Data subelement, X509Certificate element: */
        SOAPElement x509certificate = x509data.addChildElement(env.createName(
                "X509Certificate", "ds", ds));
        x509certificate.addTextNode(Base64.encode(cert.getEncoded()));
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (DiscoveryService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (DiscoveryService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
