/*
 * Classname: ResourceOffering
 * Version: 1.12
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package discovery;

import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The ResourceOffering is a class used for storing the information in each of the user's
 * service offerings. This class processes the resource offering Node and stores the values in
 * variables which should be used when accessing the services.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class ResourceOffering {

    private String offeringid;
    private String resourceid;
    private Node offering;
    private String securityMechId;

    /**
     * The constructor gets a Node with the content of the offering which it parses to find the
     * resourceID, securityMechId and the service endpoint.
     * 
     * @param offering Node The offering which is an XML Node
     */
    public ResourceOffering(Node offering) {
        offeringid = Randomizer.nextID();
        this.offering = offering;
        Node idElement = findNode("ResourceID", this.offering);
        resourceid = (idElement != null ? idElement.getTextContent() : "urn:liberty:isf:implied-resource");
        Node description = findNode("Description", this.offering);

        if (description != null) {
            Node securityMech = findNode("SecurityMechID", description);
            if (securityMech != null) securityMechId = securityMech.getTextContent();
        }
    }

    /**
     * Searches the input Node to find a Node with the name specified as a parameter and
     * returns this Node. Uses null to indicate that no Node with that name was found.
     */
    private Node findNode(String withName, Node parent) {
        NodeList children = parent.getChildNodes();
        Node current;

        for (int i = 0; i < children.getLength(); i++) {
            current = children.item(i);
            if (current.getNodeName().endsWith(withName)) {
                return current;
            } else if (current.hasChildNodes()) {
                return findNode(withName, current);
            }
        }
        return null;
    }

    /**
     * This method appends the offering to the input SOAPElement and throws an Exception if
     * something goes wrong.
     * 
     * @param appendto SOAPElement The response which the offering should be added to.
     * @throws SOAPException Thrown to show that somethings went wrong and the offering could
     *         not be added
     */
    public void getOffering(SOAPElement appendto) throws SOAPException {
        turnToSoap(offering, appendto);
    }

    /**
     * This method turn the input Node notSoap into a SOAP node by adding the information from
     * the Node into the input SOAPElement.
     */
    private void turnToSoap(Node notSoap, SOAPElement parent) throws SOAPException {
        short nodeType = notSoap.getNodeType();

        if (nodeType != Node.TEXT_NODE && notSoap.getNodeName() != null) {
            SOAPElement soap = parent.addChildElement(notSoap.getNodeName());
            if (notSoap.getNodeName().endsWith("Description")) {
                SOAPElement credentialref = soap.addChildElement("CredentialRef");
                credentialref.addTextNode(offeringid);
            }

            /* Copy the attributes: */
            NamedNodeMap attributes = notSoap.getAttributes();
            if (attributes != null) {
                for (int i = 0; i < attributes.getLength(); i++) {
                    soap.setAttribute(attributes.item(i).getNodeName(), attributes.item(i).getTextContent());
                }
            }

            /* Go through all the children of the notSoap Node: */
            NodeList children = notSoap.getChildNodes();
            
            for (int i = 0; i < children.getLength(); i++) {
                turnToSoap(children.item(i), soap);
            }
        } else if (nodeType == Node.TEXT_NODE) {
            String value = notSoap.getTextContent();
            if (value != null && value.length() > 0) {
                parent.addTextNode(value);
            }
        }
    }

    /**
     * @return String Returns the ID of the offering.
     */
    public String getOfferingId() {
        return offeringid;
    }

    /**
     * @return String Returns the resourceID.
     */
    public String getResourceId() {
        return resourceid;
    }

    /**
     * @return String The message authentication method used
     */
    public String getMessageAuthMethod() {
        /*
         * This method returns the message authentication method. Seeing as we only have one we
         * simply return the name, but this method must be implemented to return the value of
         * the securityMechId variabel if more than one method is supported.
         */
        return "Bearer";
    }
}
