/*
 * Classname: AuthTransport
 * Version: 1.17
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.microedition.io.Connector;
import javax.microedition.io.HttpConnection;
import javax.microedition.io.HttpsConnection;
import javax.microedition.io.SecurityInfo;
import javax.microedition.pki.Certificate;
import javax.microedition.pki.CertificateException;

import mymobilesite.controller.Configurations;
import mymobilesite.controller.Language;

import org.ksoap2.SoapFault;
import org.ksoap2.transport.Transport;
import org.xmlpull.v1.XmlPullParserException;

/**
 * This class is responsible for establishing the session with the Identity Provider and
 * sending the SaslMessages over this connection.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class AuthTransport extends Transport {

    private HttpsConnection session;
    private OutputStream out;
    private InputStream in;

    /**
     * Opens the session with the Identity Provider.
     * 
     * @throws MessageException If the Identity Provider cannot be found or its certificate
     *         cannot be trusted.
     */
    public AuthTransport() throws MessageException {
        try {
            /* Opening the HttpsConnection used for the authentication: */
            session = (HttpsConnection) Connector.open(Configurations.IDP_URL, Connector.READ_WRITE, true);
        } catch (SecurityException se) {
            /* The server certificate is not trusted */
            throw new MessageException(Language.ERROR_IDP_NOTTRUSTED);
        } catch (IOException e) {
            /* Could not contact the IdP */
            throw new MessageException(Language.ERROR_IDP_NOTFOUND);
        } catch (ClassCastException cce) {
            /* Wrong URL. This error means that HTTPS is not supported */
            throw new MessageException(Language.ERROR_IDP_NOTFOUND);
        } catch (IllegalArgumentException iae) {
            /* Something wrong with the URL */
            throw new MessageException(Language.ERROR_IDP_NOTFOUND);
        }
    }

    /**
     * Retrieves the server certificate and extracts the name of the Identity Provider from it.
     * 
     * @return String The name of the Identity Provider.
     * @throws MessageException If the communication was disrupted.
     */
    public String getIdpName() throws MessageException {
        try {
            /* Retrieving the server certificate and extracting the name: */
            SecurityInfo secInfo = session.getSecurityInfo();
            Certificate cert = secInfo.getServerCertificate();
            return getName(cert.getSubject());
        } catch (CertificateException ce) {
            /* Could not trust the IdP */
            throw new MessageException(Language.ERROR_IDP_NOTTRUSTED);
        } catch (IOException e) {
            /* Could not get the certificate */
            throw new MessageException(Language.ERROR_IDPNAME_NOTFOUND);
        }
    }

    /**
     * @return String Gets the part of the subject that we want the user to see.
     */
    private String getName(String subject) {
        int startOU = subject.indexOf("OU=");
        int endOU = subject.indexOf(";", startOU);

        int startO = subject.indexOf("O=");
        int endO = subject.indexOf(";", startO);

        return subject.substring(startO + 2, endO) + ", "
                + subject.substring(startOU + 3, endOU);
    }

    /**
     * Sends the message and receives the response.
     * 
     * @param request SaslMessage The SaslRequest that is to be sent and the message in which
     *        the response should be stored.
     * @throws SoapFault If the server returned a message indicating a fault.
     * @throws IOException If the communication was disrupted.
     * @throws XmlPullParserException If the response could not be parsed as XML.
     */
    public void call(SaslMessage request) throws SoapFault, IOException,
            XmlPullParserException {
        byte[] requestData = createRequestData(request);
        out = null;
        in = null;
        
        try {
            session.close();
            session = (HttpsConnection) Connector.open(Configurations.IDP_URL, Connector.READ_WRITE, true);
            session.setRequestProperty("SOAPAction", "");
            session.setRequestProperty("Content-Type", "text/xml");
            session.setRequestProperty("Content-Length", "" + requestData.length);
            session.setRequestProperty("User-Agent", "kSOAP/2.0");
            session.setRequestMethod(HttpConnection.POST);
            out = session.openOutputStream();
            out.write(requestData, 0, requestData.length);
            out.close();
            out = null;
            requestData = null;
            in = session.openInputStream();
            parseResponse(request, in);
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (Throwable e) {
                }
                in = null;
            }
        }
        if (request.bodyIn instanceof SoapFault) {
            throw ((SoapFault) request.bodyIn);
        }
    }

    /**
     * Closes the input and output streams, but keeps the session
     */
    public void reset() {
        if (in != null) {
            try {
                in.close();
            } catch (IOException e) {
            } finally {
                in = null;
            }
        }
        if (out != null) {
            try {
                out.close();
            } catch (IOException e) {
            } finally {
                out = null;
            }
        }
    }

    /**
     * Closes the session with the authentication service at the Identity Provider
     */
    public void disconnect() {
        try {
            session.close();
        } catch (IOException e) {
        } finally {
            session = null;
        }
    }
}
