/*
 * Classname: DiscoveryQuery
 * Version: 1.15
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import java.io.IOException;
import java.util.Vector;

import mymobilesite.controller.Configurations;
import mymobilesite.controller.Language;

import org.ksoap2.transport.HttpTransport;
import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

/**
 * This class is responsible for constructing and sending the discovery queries. Furthermore,
 * it processes the query responses and stores the ResourceOfferings retrieved.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class DiscoveryQuery extends LibertyMessage {

    private HttpTransport session;
    private ResourceOffering[] providers;

    /**
     * The constructor creates a DiscoveryQuery object based on the ResourceOffering.
     * 
     * @param offering ResourceOffering The ResouceOffering containing the url and token for
     *        the Discovery Service.
     */
    public DiscoveryQuery(ResourceOffering offering) {
        super();
        this.offering = offering;
    }

    /**
     * This method creates the transport that will be used in the communication with the
     * Discovery Service.
     * 
     * @return HttpTransport The transport that will be used by this class to send the query
     *         and receive the response.
     */
    public HttpTransport createTransport() {
        session = new HttpTransport(offering.getUrl());
        return session;
    }

    /**
     * This method constructs the query, sends it and processes the response.
     * 
     * @return ResourceOffering[] A list of the ResourceOfferings and the corresponding tokens
     *         returned by the Discovery Service.
     * @throws MessageException If something wrong happens, an exception with the appropriate
     *         error code is thrown.
     */
    public ResourceOffering[] sendQuery() throws MessageException {

        /* If the transport has not been created yet, create it: */
        if (session == null) session = new HttpTransport(offering.getUrl());

        /* Create the query itself first and the headers second to keep the timestamp as fresh as possible: */
        createBody();

        /* Creating the correlation header last to keep the timestamp fresh: */
        createSecurityHeader();
        createCorrelationHeader(false);
        headerOut = new Element[] { correlationHeader, securityHeader };

        try {
            /* Try to send the query: */
            session.call(offering.getSoapAction(), this);
        } catch (IOException e) {
            /* The communication was obstructed */
            throw new MessageException(Language.ERROR_DS_LOSTCONTACT);
        } catch (XmlPullParserException e) {
            /* The response could not be parsed as XML */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
        try {
            /* Try to validate the correlation header: */
            processCorrelationHeader();
        } catch (MessageException me) {
            if (me.getErrorCode() == -1) {
                /* The timestamp of the correlation header deviated too much from the current time: */
                throw new MessageException(Language.ERROR_DS_TIMESTAMPDEVIATION);
            } else {
                /* Some other exception occured: */
                throw me;
            }
        }

        /* The correlation header was ok, process the body: */
        processResponseBody("QueryResponse");

        /* Success! Return the list of ResourceOfferings: */
        return providers;
    }

    /**
     * Creates the query to be sent to the Discovery Service.
     */
    private void createBody() {
        Element query = new Element();
        query.setPrefix("disco", disco);
        query.setNamespace(disco);
        query.setName("Query");
        query.addChild(Element.ELEMENT, offering.getResourceId());
        Element requestedServiceType = query.createElement(disco, "RequestedServiceType");
        Element serviceType = requestedServiceType.createElement(disco, "ServiceType");

        /* Requesting service providers capable of interactin with this application: */
        serviceType.addChild(Element.TEXT, Configurations.SERVICE_TYPE_URI);
        requestedServiceType.addChild(Element.ELEMENT, serviceType);
        query.addChild(Element.ELEMENT, requestedServiceType);
        bodyOut = query;
    }

    /**
     * This method processes the current element and stores the resource offerings so that they
     * can be returned to the LibertyProtocol later.
     * 
     * @param current Element The element that is to be processed.
     * @throws MessageException If there are no results or the message is invalid.
     */
    protected void processResponse(Element current) throws MessageException {
        Element child;
        String statusCode = null;
        /* Try to find the Credentials element: */
        Element credentials = null;
        try {
            credentials = current.getElement(disco, "Credentials");
        } catch (Exception e) {}

        /* Create a dynamic list in case not all resource offerings are valid: */
        Vector offerings = new Vector(current.getChildCount());
        boolean hasStatus = false;

        /* Go through the current element's children: */
        for (int i = 0; i < current.getChildCount(); i++) {
            child = current.getElement(i);
            if (child == null) {
                /* Do nothing */
            } else if (child.getName() == null) {
                /* Do nothing */
            } else if (child.getName().equals("Status")) {
                /* This is the status element */
                hasStatus = true;
                statusCode = child.getText(0);
                if (statusCode != null) {
                    if (statusCode.toLowerCase().indexOf("failed") > -1) {
                        /* The result is not ok, don't keep it: */
                        throw new MessageException(Language.ERROR_DS_NORESULTS);
                    }
                }
            } else if (child.getName().equals("ResourceOffering")) {
                /* This is a resource offering */
                try {
                    /* Try to add it to the list: */
                    offerings.addElement(new ResourceOffering(child, credentials));
                } catch (MessageException me) {
                    /* This resource offering is not valid, skip it. */
                }
            }
        }
        providers = new ResourceOffering[offerings.size()];

        /* Move the ResourceOfferings to the array of ResourceOfferings */
        offerings.copyInto(providers);
        if (providers.length == 0) {
            /* There are no results */
            throw new MessageException(Language.ERROR_DS_NORESULTS);
        } else if (!hasStatus || statusCode == null) {
            /* The status element was missing */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
    }
}
