/*
 * Classname: LibertyMessage
 * Version: 1.16
 * Date: 22.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import java.util.Date;

import mymobilesite.controller.Configurations;
import mymobilesite.controller.Language;

import org.kobjects.isodate.IsoDate;
import org.ksoap2.SoapEnvelope;
import org.kxml2.kdom.Element;
import org.kxml2.kdom.Node;

/**
 * This abstract class is the general purpose superclass of the specific Liberty messages sent
 * from the mobile phone to the Identity Provider, Discovery Service and Web Service Provider.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public abstract class LibertyMessage extends SoapEnvelope {

    /** The namespace for the Liberty SOAP Binding. */
    protected static final String sb = "urn:liberty:sb:2003-08";
    /** The namespace for the Liberty Authentication Service. */
    protected static final String sa = "urn:liberty:sa:2004-04";
    /** The namespace for the WS-Security header. */
    protected static final String wsse = "http://schemas.xmlsoap.org/ws/2002/04/secext";
    /** The namespace for the Liberty Discovery Service. */
    protected static final String disco = "urn:liberty:disco:2003-08";

    protected String sentMessageID;
    protected String receivedMessageID;
    protected Element correlationHeader;
    protected Element securityHeader;

    protected ResourceOffering offering;

    /**
     * The constructor creates the SoapEnvelope with the correct SOAP version.
     */
    protected LibertyMessage() {
        super(SoapEnvelope.VER11);
    }

    /**
     * Creates the Liberty SOAP Binding correlation header.
     * 
     * @param second boolean If this parameter is true it means that this message is sent in
     *        response to a previously received message.
     */
    protected void createCorrelationHeader(boolean second) {
        correlationHeader = new Element();
        correlationHeader.setPrefix("sb", sb);
        correlationHeader.setNamespace(sb);
        correlationHeader.setName("Correlation");
        correlationHeader.setAttribute(env, "mustUnderstand", "1");
        sentMessageID = Randomizer.nextID();
        correlationHeader.setAttribute(sb, "messageID", sentMessageID);

        if (second) {
            /* This request is a response to a previously received response */
            correlationHeader.setAttribute(sb, "refToMessageID", receivedMessageID);
        }
        correlationHeader.setAttribute(sb, "timestamp", IsoDate.dateToString(new Date(System
                .currentTimeMillis()), IsoDate.DATE_TIME));
    }

    /**
     * Creates the WS-Security security header that contains the token specified in the
     * ResourceOffering. This method is not used by the SaslMessage class.
     */
    protected void createSecurityHeader() {
        securityHeader = new Element();
        securityHeader.setPrefix("wsse", wsse);
        securityHeader.setNamespace(wsse);
        securityHeader.setName("Security");
        securityHeader.addChild(Element.ELEMENT, offering.getToken());
    }

    /**
     * This method checks that the received message's correlation header is valid and throws 
     * MessageException if there is something wrong with the header.
     */
    protected void processCorrelationHeader() throws MessageException {

        /* If there are no headers in the return message it means that the correlation header does not exist: */
        if (headerIn == null) {
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
        boolean correlation = false;

        /* Going through the headers: */
        for (int i = 0; i < headerIn.length; i++) {
            if (headerIn[i].getName().equals("Correlation")) {
                /* This is the correlation header */
                correlation = true;
                String attName;
                String attValue;
                boolean mID = false;
                boolean timest = false;

                /* Going through the attributes of the correlation header: */
                for (int j = 0; j < headerIn[i].getAttributeCount(); j++) {
                    attName = headerIn[i].getAttributeName(j);
                    attValue = headerIn[i].getAttributeValue(j);

                    if (attValue == null) {
                        /* Something wrong has happend with the content of this message */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    } else if (attName.equals("messageID")) {
                        if (receivedMessageID != null && attValue.equals(receivedMessageID)) {
                            /* We have received this message before, don't process! */
                            throw new MessageException(Language.ERROR_REPLAY_DETECTED);
                        } else {
                            /* This message contains a valid message id */
                            mID = true;
                            receivedMessageID = attValue;
                        }
                    } else if (attName.equals("timestamp") && attValue != null) {
                        long timestamp = IsoDate.stringToDate(attValue, IsoDate.DATE_TIME) .getTime();
                        long now = System.currentTimeMillis();

                        if (Math.abs(now - timestamp) > Configurations.TIMESTAMP_DEVIATION_MILLIS) {
                            /* The timestamp of the message is too different from the current time */
                            throw new MessageException(-1, "");
                        } else {
                            /* The timestamp has been checked and is ok */
                            timest = true;
                        }
                    } else if (attName.equals("refToMessageID")) {
                        /* The response has a reference to a previously sent message */

                        if (!attValue.equals(sentMessageID)) {
                            /* The response belongs to another request */
                            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                        }
                    } else if (attName.equals("mustUnderstand")
                            && !(attValue.equals("1") || attValue.equals("0"))) {
                        /* The mustUnderstand attribute was present, but its value was false */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    }
                }
                if (!(mID && timest)) {
                    /* The correlation header did not contain the required attributes */
                    throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                }
            }
        }
        if (!correlation) {
            /* The correlation header was missing */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
        /* If the method reaches this point it means that the header is ok */
    }

    /**
     * This method looks for an Element in the response body and starts the subclass' specific
     * response processing if the Element is found. If it cannot find the Element, an exception
     * is thrown.
     * 
     * @param lookFor String The local name of the Element to look for.
     * @throws MessageException If the Element couldn't be found or if the response body is
     *         invalid.
     */
    protected void processResponseBody(String lookFor) throws MessageException {
        if (bodyIn == null || !(bodyIn instanceof Node)) {
            /* There is something wrong with the message body */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }

        Node bodyNode = (Node) bodyIn;
        if (bodyNode.getChildCount() <= 0) {
            /* There is something wrong with the message body */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }

        Element response;
        boolean foundIt = false;

        for (int j = 0; j < bodyNode.getChildCount(); j++) {
            response = bodyNode.getElement(j);

            if (response == null) {
                /* Do nothing */
            } else if (response.getName() == null) {
                /* Do nothing */
            } else if (response.getName().endsWith(lookFor)) {
                /* Found it: */
                foundIt = true;
                processResponse(response);
            } else {
                Element child;

                for (int i = 0; i < response.getChildCount(); i++) {
                    child = response.getElement(i);

                    if (child == null) {
                        /* Do nothing */
                    } else if (child.getName() == null) {
                        /* Do nothing */
                    } else if (child.getName().endsWith(lookFor)) {
                        foundIt = true;
                        /* Found it: */
                        processResponse(child);
                    }
                }
            }
        }
        if (!foundIt) {
            /* The element is missing */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
    }

    /**
     * This method processes the current element according to the rules of the implementing
     * subclass.
     * 
     * @param current Element The element that is to be processed.
     * @throws MessageException If something wrong happens.
     */
    protected abstract void processResponse(Element current) throws MessageException;

}
