/*
 * Classname: LibertyProtocol
 * Version: 1.21
 * Date: 21.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import mymobilesite.controller.Language;
import mymobilesite.controller.LibertyProtocolListener;
import mymobilesite.controller.Service;

import org.ksoap2.transport.HttpTransport;

/**
 * The LibertyProtocol class is responsible for coordinating the messages that are sent between
 * the mobile phone and the Identity Provdier, Discovery Service and Web Service Provider.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class LibertyProtocol implements Runnable {

    private AuthTransport idpSession;
    private HttpTransport discoSession;
    private HttpTransport providerSession;
    private ResourceOffering discoResource;
    private ResourceOffering[] resourceOfferings;
    private LibertyProtocolListener listener;
    private Thread executingThread;

    /** State constant indicating that the class has not yet started its work. */
    public static final int NOT_STARTED = -1;
    
    /** State constant indicating that the class is establishing the session with the Identity Provider. */
    public static final int CONTACTING_IDP = 0;
    
    /** State constant indicating that the class is in the process of authenticating the user. */
    public static final int AUTHENTICATING_USER = 1;
    
    /** State constant indicating that the class is retrieving the list of resource offerings from the 
     * Discovery Service. */
    public static final int RETRIEVING_PROVIDER_LIST = 2;
    
    /** State constant indicating that the class is requesting the service chosen by the user from the 
     * Web Service Provider. */
    public static final int REQUESTING_SERVICE = 3;
    
    /** State constant indicating that the class has completed its tasks successfully. */
    public static final int DONE = 4;

    private int currentState;
    private boolean abort;
    private boolean running;
    private String un;
    private String pwd;
    private int userchoice;

    /**
     * The contstructor registers the listener and initialises the executing thread.
     * 
     * @param listener LibertyProtocolListener The listener that should be notified of events
     *        in this class.
     */
    public LibertyProtocol(LibertyProtocolListener listener) {
        this.listener = listener;
        executingThread = new Thread(this);
        currentState = NOT_STARTED;
        abort = false;
    }

    /**
     * This method can be called if the current operation of the LibertyProtocol class needs to
     * be aborted for some reason. It attempts to halt the execution by closing the input and
     * output streams, which will cause the communication to fail and throw an exception.
     */
    public void abort() {
        switch (currentState) {
        case AUTHENTICATING_USER:
            abort = true;
            if (idpSession != null) idpSession.reset();
            break;
        case RETRIEVING_PROVIDER_LIST:
            abort = true;
            if (discoSession != null) discoSession.reset();
            break;
        case REQUESTING_SERVICE:
            abort = true;
            if (providerSession != null) providerSession.reset();
            break;
        default:
            abort = true;
            break;
        }
    }

    /**
     * This method causes the class to enter the CONTACTING_IDP state and starts the executing
     * thread.
     */
    public void establishIdpSession() {
        abort = false;
        running = true;
        currentState = CONTACTING_IDP;
        if (!executingThread.isAlive()) {
            executingThread = new Thread(this);
            executingThread.start();
        }
    }

    /**
     * This method causes the class to enter the AUTHENTICATING_USER state and starts the
     * executing thread. If the authentication is succsessful the state will change to
     * RETRIEVING_PROVIDER_LIST automatically.
     * 
     * @param un String The username to be authenticated.
     * @param pwd String The password provided by the user.
     */
    public void getProviders(String un, String pwd) {
        abort = false;
        this.un = un;
        this.pwd = pwd;
        currentState = AUTHENTICATING_USER;
        running = true;
        if (!executingThread.isAlive()) {
            executingThread = new Thread(this);
            executingThread.start();
        }
    }

    /**
     * In some cases the Controller will have to retry the authentication. If the credentials
     * are the same as last time, it can call this method instead of the getProviders(un, pwd)
     * method.
     */
    public void retryGetProviders() {
        getProviders(un, pwd);
    }

    /**
     * In some cases the Controller will have to retry to retrieve the list of providers. If
     * the authentication is valid, it can use this method to contact the Discovery Service
     * once more. The class will then enter the RETRIEVING_PROVIDER_LIST directly.
     */
    public void retryDiscoQuery() {
        abort = false;
        running = true;
        currentState = RETRIEVING_PROVIDER_LIST;
        if (!executingThread.isAlive()) {
            executingThread = new Thread(this);
            executingThread.start();
        }
    }

    /**
     * This method causes the class to enter the REQUESTING_SERVICE state and starts the
     * executing thread. It will then try to contact the Web Service Provider chosen by the
     * user.
     * 
     * @param userchoice int The index of the WSP chosen by the user.
     */
    public void requestService(int userchoice) {
        abort = false;
        running = true;
        currentState = REQUESTING_SERVICE;
        this.userchoice = userchoice;
        if (!executingThread.isAlive()) {
            executingThread = new Thread(this);
            executingThread.start();
        }
    }

    /**
     * This method takes care of the actual execution and performs tasks appropriate to the
     * current state of the class. When the results are ready it notifies the listener by
     * calling the resultsReady method. If it encounters an exception that was not caused by an
     * abort, it notifies the listener with the encounteredException method.
     * 
     * @see java.lang.Runnable#run()
     */
    public void run() {
        while (running) {
            switch (currentState) {
            case CONTACTING_IDP:
                running = false;
                try {
                    idpSession = new AuthTransport();
                    /* The opening of the session was successful, deliver the name to the listener: */
                    listener.resultsReady(currentState, idpSession.getIdpName());
                } catch (MessageException me) {
                    if (idpSession != null) {
                        idpSession.reset();
                        idpSession.disconnect();
                        idpSession = null;
                    }
                    /* Something is wrong, notify the listener: */
                    listener.encounteredException(currentState, me);
                    currentState = NOT_STARTED;
                }
                break;
            case AUTHENTICATING_USER:
                SaslMessage sasl = new SaslMessage();
                try {
                    /* Try to get the ResourceOffering for the DS: */
                    discoResource = sasl.sendRequest(un, pwd, idpSession);
                    if (!abort) {
                        /* Success! Notify the listener: */
                        listener.resultsReady(currentState, discoResource);
                        /* Go to next state: */
                        currentState = RETRIEVING_PROVIDER_LIST;
                    }
                } catch (MessageException me1) {
                    running = false;

                    /* Something wrong happened, check if the operation was aborted: */
                    if (!abort) {
                        /*
                         * There was no abort, reset the session and check if we should try
                         * again:
                         */
                        idpSession.reset();
                        if (me1.getErrorCode() == Language.ERROR_INVALID_MESSAGE) {
                            /* Invalid message. Try again: */
                            sasl = new SaslMessage();
                            try {
                                discoResource = sasl.sendRequest(un, pwd, idpSession);
                                listener.resultsReady(currentState, discoResource);
                                currentState = RETRIEVING_PROVIDER_LIST;
                                running = true;
                            } catch (MessageException me2) {
                                running = false;
                                /* Something wrong happened again. Was the operation aborted? */
                                if (!abort) {
                                    /* The operation was not aborted. Notify the listener of the exception: */
                                    idpSession.reset();
                                    listener.encounteredException(currentState, me2);
                                }
                                currentState = CONTACTING_IDP;
                            }
                        } else {
                            /* No point in trying again, notify the listner straight away: */
                            listener.encounteredException(currentState, me1);
                            currentState = CONTACTING_IDP;
                        }
                    } else {
                        /* The exception was caused by an abort. Don't report the exception. */
                        currentState = CONTACTING_IDP;
                    }
                }
                break;
            case RETRIEVING_PROVIDER_LIST:
                running = false;
                DiscoveryQuery disco = new DiscoveryQuery(discoResource);
                discoSession = disco.createTransport();
                try {
                    /* Try to get the list of ResourceOfferings: */
                    resourceOfferings = disco.sendQuery();

                    /* Success! Retrieve the names of the providers: */
                    String[] names = new String[resourceOfferings.length];

                    for (int i = 0; i < names.length; i++) {
                        names[i] = resourceOfferings[i].getName();
                    }
                    if (!abort) {
                        /* Done. Notify the listener: */
                        listener.resultsReady(currentState, names);

                        /* Close the idpSession. We don't need it any more: */
                        if (idpSession != null) {
                            idpSession.disconnect();
                            idpSession = null;
                        }
                    }
                } catch (MessageException me1) {
                    /* Something wrong happened. Was the operation aborted? */
                    if (!abort) {
                        /* There was no abort, check if we should try again: */
                        if (me1.getErrorCode() == Language.ERROR_REPLAY_DETECTED
                                || me1.getErrorCode() == Language.ERROR_INVALID_MESSAGE) {
                            /* Invalid message or suspicion of replay. Try again: */
                            disco = new DiscoveryQuery(discoResource);
                            discoSession = disco.createTransport();

                            try {
                                resourceOfferings = disco.sendQuery();
                                String[] names = new String[resourceOfferings.length];

                                for (int i = 0; i < names.length; i++) {
                                    names[i] = resourceOfferings[i].getName();
                                }
                                listener.resultsReady(currentState, names);

                                if (idpSession != null) {
                                    idpSession.disconnect();
                                    idpSession = null;
                                }
                            } catch (MessageException me2) {
                                /* Something wrong happened again. Was the operation aborted? */
                                if (!abort) {
                                    /* The operation was not aborted. Notify the listener of the exception: */
                                    listener.encounteredException(currentState, me2);
                                    currentState = AUTHENTICATING_USER;

                                    /* Check if we should disconnect the IdP: */
                                    if (idpSession != null
                                            && me1.getErrorCode() != Language.ERROR_DS_NORESULTS) {
                                        idpSession.disconnect();
                                        idpSession = null;
                                    }
                                } else {
                                    /* The exception was caused by an abort. Don't report the exception. */
                                    currentState = AUTHENTICATING_USER;
                                }
                            }
                        } else {
                            /* No point in trying again, notify the listner straight away: */
                            listener.encounteredException(currentState, me1);
                            currentState = AUTHENTICATING_USER;

                            /* Check if we should disconnect the IdP: */
                            if (idpSession != null
                                    && me1.getErrorCode() != Language.ERROR_DS_NORESULTS) {
                                idpSession.disconnect();
                                idpSession = null;
                            }
                        }
                    } else {
                        /* The exception was caused by an abort. Don't report it. */
                        currentState = AUTHENTICATING_USER;
                    }
                } finally {
                    if (discoSession != null) {
                        discoSession.reset();
                        discoSession = null;
                    }
                }
                break;
            case REQUESTING_SERVICE:
                running = false;
                WspRequest wsp = new WspRequest(resourceOfferings[userchoice]);
                providerSession = wsp.createTransport();
                try {
                    /* Try to retrieve the services: */
                    Service[] services = wsp.sendRequest();
                    if (!abort) {
                        /* Success! Notify the listener: */
                        listener.resultsReady(currentState, services);
                        /* This class' work is done: */
                        currentState = DONE;
                    }
                } catch (MessageException me1) {
                    /* Something wrong happened. Was the operation aborted? */
                    if (!abort) {
                        /* There was no abort, check if we should try again: */
                        if (me1.getErrorCode() == Language.ERROR_REPLAY_DETECTED
                                || me1.getErrorCode() == Language.ERROR_INVALID_MESSAGE) {
                            /* Invalid message or suspicion of replay. Try again: */
                            wsp = new WspRequest(resourceOfferings[userchoice]);
                            providerSession = wsp.createTransport();
                            try {
                                Service[] services = wsp.sendRequest();
                                listener.resultsReady(currentState, services);
                                currentState = DONE;
                            } catch (MessageException me2) {
                                /* Something wrong happened again. Was the operation aborted? */
                                if (!abort) {
                                    /* The operation was not aborted. Notify the listener of the exception: */
                                    listener.encounteredException(currentState, me2);
                                    currentState = RETRIEVING_PROVIDER_LIST;
                                } else {
                                    /* The exception was caused by an abort. Don't report it. */
                                    currentState = RETRIEVING_PROVIDER_LIST;
                                }
                            }
                        } else {
                            /* No point in trying again, notify the listner straight away: */
                            listener.encounteredException(currentState, me1);
                            currentState = RETRIEVING_PROVIDER_LIST;
                        }
                    } else {
                        /* The exception was caused by an abort. Don't report it. */
                        currentState = RETRIEVING_PROVIDER_LIST;
                    }
                } finally {
                    if (providerSession != null) {
                        providerSession.reset();
                        providerSession = null;
                    }
                }
                break;
            default:
                running = false;
                break;
            }
        }
    }
}
