/*
 * Classname: ResourceOffering
 * Version: 1.15
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import mymobilesite.controller.Configurations;
import mymobilesite.controller.Language;

import org.kxml2.kdom.Element;

/**
 * This class is responsible for parsing the contents of a ResourceOffering element and storing
 * the relevant information so that it can be used to contact the service provider.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class ResourceOffering {

    private Element resourceid;
    private String name;
    private String url;
    private Element token;
    private String credentialReference;
    private String saml = "urn:oasis:names:tc:SAML:1.0:assertion";
    private String soapaction;

    /**
     * The constructor starts the process of parsing the request and retrieving the appropriate
     * token from the credentials.
     * 
     * @param offering Element An Element containing the resource offering.
     * @param credentials Element An element containing a lot of tokens as its subelements.
     * @throws MessageException If any of the parameters are invalid.
     */
    public ResourceOffering(Element offering, Element credentials) throws MessageException {
        processOffering(offering);
        if (url == null) throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        if (credentials != null) processCredentials(credentials);
        if (token != null && token.getNamespace() == null) {
            token.setPrefix("saml", saml);
            token.setNamespace(saml);
        }
    }

    /**
     * Processes the Element containing the resource offering and stores the values that are
     * needed to contact the service provider later.
     * 
     * @param offering Element An Element containing the resource offering.
     */
    private void processOffering(Element offering) {
        Element child;
        for (int i = 0; i < offering.getChildCount(); i++) {
            child = offering.getElement(i);
            if (child != null) {
                if (child.getName().endsWith("Abstract")) {
                    /* This element contains the name of the service */
                    StringBuffer nameBuffer = new StringBuffer();

                    for (int j = 0; j < child.getChildCount(); j++) {
                        nameBuffer.append(child.getText(j));
                    }
                    String nameText = nameBuffer.toString();

                    if (nameText == null && url != null) {
                        /* No name is specified, display the url instead */
                        name = (url.length() > Configurations.NAME_LENGTH_CONSTRAINT ? 
                                url.substring(0, Configurations.NAME_LENGTH_CONSTRAINT) : url);
                    } else {
                        int nameBeginIndex = 0;
                        if (nameText.indexOf(";;;" + Configurations.LANGUAGE + ";;;") > -1) {
                            /* A name in the preferred language exists */
                            nameBeginIndex = nameText.indexOf(";;;" + Configurations.LANGUAGE + ";;;");
                            nameBeginIndex += (";;;" + Configurations.LANGUAGE + ";;;").length();
                        } else if (nameText.indexOf(";;;gb;;;") > -1) {
                            /* A name in the default language exists */
                            nameBeginIndex = nameText.indexOf(";;;gb;;;");
                            nameBeginIndex += ";;;gb;;;".length();
                        } else if (nameText.indexOf(";;;", 3) > -1) {
                            /* Choose the first name */
                            nameBeginIndex = nameText.indexOf(";;;", 3) + ";;;".length();
                        }
                        int nameEndIndex = nameText.indexOf(";;;", nameBeginIndex);
                        if (nameEndIndex < 0) nameEndIndex = nameText.length();
                        nameText = nameText.substring(nameBeginIndex, nameEndIndex);
                        name = (nameText.length() > Configurations.NAME_LENGTH_CONSTRAINT ? 
                                nameText.substring(0, Configurations.NAME_LENGTH_CONSTRAINT) : nameText);
                    }
                } else if (child.getName().endsWith("ServiceInstance")) {
                    /* Process this elements subelements */
                    processOffering(child);
                } else if (child.getName().endsWith("ResourceID")) {
                    /* This element contains the ResourceID */
                    resourceid = child;
                } else if (child.getName().endsWith("EncryptedResourceID")) {
                    /* This element contains the EcryptedResourceID, which can be used as a ResourceID */
                    resourceid = child;
                } else if (child.getName().endsWith("Description")) {
                    /* Process this elements subelements */
                    processOffering(child);
                } else if (child.getName().endsWith("Endpoint")) {
                    /* This element contains the url */
                    url = child.getText(0);
                    if (name == null) {
                        name = (url.length() > Configurations.NAME_LENGTH_CONSTRAINT ? 
                                url.substring(0, Configurations.NAME_LENGTH_CONSTRAINT) : url);
                    }
                } else if (child.getName().endsWith("SoapAction")) {
                    /* This element contains the SoapAction */
                    soapaction = child.getText(0);
                } else if (child.getName().endsWith("CredentialRef")) {
                    /* This element contains the reference to the token */
                    credentialReference = child.getText(0);
                } else if (child.getName().endsWith("SecurityMechID")) {
                    /* Something must be done here to support several mechanisms */
                }
            }
        }
    }

    /**
     * This method goes through the credentials and stores the appropriate one, so that it can
     * be used to access the service it is supposed to be used for.
     * 
     * @param credentials Element An element containing a lot of tokens as its subelements.
     * @throws MessageException If there is something wrong with the credentials.
     */
    private void processCredentials(Element credentials) throws MessageException {
        if (credentialReference == null && credentials.getChildCount() > 1) {
            /*
             * The credentials element contains more than one token, but the reference is null
             * and we cannot know which one to choose
             */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        } else if (credentialReference == null && credentials.getChildCount() == 1) {
            /* The reference is null, but there is only one token */
            token = credentials.getElement(0);
            return;
        }
        Element child;

        /* Going through the tokens: */
        for (int i = 0; i < credentials.getChildCount(); i++) {
            child = credentials.getElement(i);
            if (child == null) {
                /* Do nothing */
            } else if (child.getName() == null) {
                /* Do nothing */
            } else if (child.getName().endsWith("Assertion")) {
                String id = child.getAttributeValue(child.getNamespace(), "AssertionID");
                if (id != null) {
                    if (id.equals(credentialReference)) {
                        /* This is the correct token */
                        token = child;
                        return;
                    }
                }
            }
        }
        /* If we get this far it means that we have processed all the tokens without finding the correct one */
        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
    }

    /**
     * @return String The name of the service provider.
     */
    public String getName() {
        return name;
    }

    /**
     * @return String The url pointing to the service provider.
     */
    public String getUrl() {
        return url;
    }

    /**
     * @return Element The token that can be used to access the service provider.
     */
    public Element getToken() {
        return token;
    }

    /**
     * @return String The ResourceID or EncryptedResourceID of this offering
     */
    public Element getResourceId() {
        return resourceid;
    }

    /**
     * @return String The SoapAction of this offering
     */
    public String getSoapAction() {
        return (soapaction != null ? soapaction : "\"\"");
    }
}
