/*
 * Classname: SaslMessage
 * Version: 1.18
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import java.io.IOException;

import mymobilesite.controller.Language;

import org.ksoap2.SoapFault;
import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

/**
 * This class is responsible for handling the communication with the authentication service.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class SaslMessage extends LibertyMessage {

    private static final char UPPERCASE = 'u';
    private static final char LOWERCASE = 'l';
    private static final char SELECT = 's';
    private static final char TRUNCATION = 't';

    private String[][] transformations;

    private Element request;
    private String un;

    private ResourceOffering disco;
    private boolean needAnotherRequest = false;

    /**
     * The constructor creates an empty SaslMessage.
     */
    public SaslMessage() {
        super();
    }

    /**
     * This method constructs the SaslRequest, sends it and processes the response. If
     * everything is successful it returns the ResourceOffering for the Discovery Service.
     * 
     * @param un String The username to be authenticated.
     * @param pwd String The password provided by the user.
     * @param session AuthTransport The transport to be used in the communication with the
     *        authentication service.
     * @return ResourceOffering The ResourceOffering for the DiscoveryService.
     * @throws MessageException If something wrong happens or the authentication fails, an
     *         exception with the apporpriate error code is thrown.
     */
    public ResourceOffering sendRequest(String un, String pwd, AuthTransport session)
            throws MessageException {
        this.un = un;
        createRequest(pwd, false);
        
        try {
            /* Sending the request and receiving the response: */
            session.call(this);
        } catch (SoapFault e) {
            /* The IdP returned a fault */
            throw new MessageException(Language.ERROR_IDP_FAULT);
        } catch (IOException e) {
            /* Something wrong happened in the communication */
            throw new MessageException(Language.ERROR_IDP_LOSTCONTACT);
        } catch (XmlPullParserException e) {
            /* The response from the server could not be parsed as XML */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
        /* Checking that the response header is ok: */
        try {
            processCorrelationHeader();
        } catch (MessageException me) {
            if (me.getErrorCode() == -1) {
                throw new MessageException(Language.ERROR_IDP_TIMESTAMPDEVIATION);
            } else {
                throw me;
            }
        }
        processResponseBody("SASLResponse");
        
        if (needAnotherRequest) {
            String tmp = transform(pwd);
            if (!tmp.equals(pwd)) {
                createRequest(pwd, true);
                try {
                    /* Sending the request and receiving the response: */
                    session.call(this);
                } catch (SoapFault e) {
                    /* The IdP returned a fault */
                    throw new MessageException(Language.ERROR_IDP_FAULT);
                } catch (IOException e) {
                    /* Something wrong happened in the communication */
                    throw new MessageException(Language.ERROR_IDP_LOSTCONTACT);
                } catch (XmlPullParserException e) {
                    /* The response from the server could not be parsed as XML */
                    throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                }
                try {
                    processCorrelationHeader();
                } catch (MessageException me) {
                    if (me.getErrorCode() == -1) {
                        throw new MessageException(Language.ERROR_IDP_TIMESTAMPDEVIATION);
                    } else {
                        throw me;
                    }
                }
                processResponseBody("SASLResponse");
            } else {
                /* There is no point in continuing if the password is the same as before */
                throw new MessageException(Language.ERROR_AUTH_NOTACCEPTED);
            }
        }
        return disco;
    }

    /**
     * This method creates the request that is to be set to the authentication service at the
     * Identity Provider.
     * 
     * @param pwd String The password provided by the user.
     * @param second boolean If this parameter is true it means that this message is sent in
     *        response to a previously received message.
     */
    private void createRequest(String pwd, boolean second) {
        /* Creating the SASLRequest body: */
        request = new Element().createElement(sa, "SASLRequest");
        request.setPrefix("sa", sa);
        request.setNamespace(sa);
        request.setName("SASLRequest");
        request.setAttribute(sa, "mechanism", "PLAIN");
        request.setAttribute(sa, "authzID", un);

        /* Adding the response in the first request: */
        Element data = request.createElement(sa, "Data");
        data.addChild(Element.TEXT, "%x00" + un + "%x00" + pwd);
        request.addChild(Element.ELEMENT, data);
        bodyOut = request;

        /* Creating the correlation header last to keep the timestamp fresh: */
        createCorrelationHeader(second);
        headerOut = new Element[] { correlationHeader };
    }

    /**
     * This method processes the current element and stores the ResourceOffering for the
     * Discovery Service so that it can be returned to the LibertyProtocol later.
     * 
     * @param current Element The element that is to be processed.
     * @throws MessageException If the authentication is rejected or the message is invalid.
     */
    protected void processResponse(Element element) throws MessageException {
        String attName;
        String attValue;
        boolean hasMechanism = false;

        /* Going through the attributes of the SASLResponse: */
        for (int j = 0; j < element.getAttributeCount(); j++) {
            attName = element.getAttributeName(j);
            attValue = element.getAttributeValue(j);

            if (attValue == null) {
                /* Something wrong has happend with the content of this message */
                throw new MessageException(Language.ERROR_INVALID_MESSAGE);
            } else if (attName.endsWith("serverMechanism")) {
                if (!(attValue.toUpperCase().indexOf("PLAIN") > -1)) {
                    /* The server suggested an unsupported authentication mechanism */
                    throw new MessageException(Language.ERROR_AUTH_NOTACCEPTED);
                } else {
                    /* The server mechanism is ok */
                    hasMechanism = true;
                }
            }
        }

        if (element.getAttributeCount() > 0 && !hasMechanism) {
            /* This response element does not contain the requred attributes */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }

        Element child;
        boolean hasStatus = false;
        boolean hasStatusCode = false;
        boolean hasDiscoResource = false;

        Element resourceOffering = null;
        Element credentials = null;

        for (int i = 0; i < element.getChildCount(); i++) {
            child = element.getElement(i);
            if (child != null) {
                if (child.getName().equals("Status")) {
                    hasStatus = true;
                    String statusCode = child.getText(0);
                    if (statusCode == null) {
                        /* Something wrong has happend with the content of this message */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    } else {
                        hasStatusCode = true;
                        if (statusCode.toLowerCase().endsWith("abort")) {
                            /* The server has aborted the authentication attempt */
                            throw new MessageException(Language.ERROR_AUTH_NOTACCEPTED);
                        } else if (statusCode.toLowerCase().endsWith("continue")) {
                            /*
                             * The server is not satisfied with the results so far, we need to
                             * send another request
                             */
                            needAnotherRequest = true;
                        } else if (statusCode.toLowerCase().endsWith("ok")) {
                            /* The authentication was successful */
                        } else {
                            /* The code attribute contains an invalid status code */
                            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                        }
                    }
                } else if (child.getName().equals("PasswordTransforms")) {
                    /*
                     * This element contains instructions for preprocessing of the password
                     * before sending it
                     */
                    Element transform;
                    Element parameter;
                    String methodUri;
                    transformations = new String[child.getChildCount()][2];

                    for (int j = 0; j < child.getChildCount(); j++) {
                        transform = child.getElement(j);
                        if (transform != null && transform.getName().equals("Transform")) {
                            methodUri = transform.getAttributeValue(transform.getNamespace(), "name");
                            /* Get the locally unique identifier of the transformation method: */
                            transformations[j][0] = methodUri.substring(
                                    methodUri.lastIndexOf(':') + 1,
                                    methodUri.lastIndexOf(':') + 2).toLowerCase();
                            for (int x = 0; x < transform.getChildCount(); x++) {
                                parameter = transform.getElement(x);
                                if (parameter != null) {
                                    if (parameter.getName().equals("Parameter")) {
                                        transformations[j][1] = parameter.getText(0);
                                    }
                                }
                            }
                        }
                    }
                } else if (child.getName().equals("ResourceOffering")) {
                    try {
                        Element serviceInst = child.getElement(child.getNamespace(), "ServiceInstance");
                        Element serviceType = serviceInst.getElement(child.getNamespace(), "ServiceType");
                        if (serviceType.getText(0).indexOf("urn:liberty:disco") > -1) {
                            resourceOffering = child;
                            if (credentials != null && disco == null) {
                                /* Both the resource offering and the credentials have been found */
                                disco = new ResourceOffering(child, credentials);
                            }
                        } else {
                            /* 
                             * This resource offering element does not contain a reference to the 
                             * discovery service
                             */
                            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                        }
                    } catch (NullPointerException npe) {
                        /* This resource offering element does not contain the requred elements */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    }
                } else if (child.getName().equals("Credentials")) {
                    /* The element contains the credentials needed to access the discovery service */
                    credentials = child;
                    if (resourceOffering != null && disco == null) {
                        /* Both the resource offering and the credentials have been found */
                        disco = new ResourceOffering(resourceOffering, credentials);
                    }
                }
            }
        }
        if (!(hasStatus || hasStatusCode || hasDiscoResource)) {
            /* This response element does not contain the requred elements */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
    }

    /**
     * In some cases the authentication service may specify that the password provided by the
     * user should be transformed in some way. This method takes care of these transformation
     * by transforming the text according to the methods specified in the transformations
     * variable.
     * 
     * @param text String The text that should be transformed.
     * @return String The transformed text.
     * @throws MessageException If one of the methods are invalid it means that the message
     *         they arrived in is invalid. Consequently a MessageException with an
     *         appropriate error code is thrown.
     */
    private String transform(String text) throws MessageException {
        char method;
        if (transformations != null) {
            for (int i = 0; i < transformations.length; i++) {
                if (transformations[i][0] != null) {
                    method = transformations[i][0].charAt(0);
                    switch (method) {
                    case UPPERCASE:
                        /* Convert the text to uppercase: */
                        text = text.toUpperCase();
                        break;
                    case LOWERCASE:
                        /* Convert the text to lowercase: */
                        text = text.toLowerCase();
                        break;
                    case SELECT:
                        /* Remove all characters that are not included in the parameter: */
                        if (transformations[i][1] != null) {

                            for (int j = 0; j < text.length(); j++) {
                                /* Check if the current character is allowed: */
                                if (transformations[i][1].indexOf(text.charAt(j)) < 0) {
                                    /*
                                     * The current character is not allowed and must be
                                     * removed:
                                     */
                                    text = text.substring(0, j) + text.substring(j + 1, text.length());
                                    j--;
                                }
                            }
                        } else {
                            /* The method requires a parameter, but didn't receive one */
                            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                        }
                        break;
                    case TRUNCATION:
                        /* Truncate the text to the given length: */
                        if (transformations[i][1] != null) {
                            try {
                                int length = Integer.parseInt(transformations[i][1]);
                                if (length < text.length()) text = text.substring(0, length);
                            } catch (NumberFormatException nfe) {
                                /* The parameter given was invalid */
                                throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                            }
                        } else {
                            /* The method requires a parameter, but didn't receive one */
                            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                        }
                        break;
                    default:
                        /* Unknown method */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    }
                }
            }
        }
        return text;
    }
}
