/*
 * Classname: WspRequest
 * Version: 1.17
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.communication;

import java.io.IOException;
import java.util.Vector;

import mymobilesite.controller.Configurations;
import mymobilesite.controller.Language;
import mymobilesite.controller.Service;

import org.ksoap2.transport.HttpTransport;
import org.kxml2.kdom.Element;
import org.xmlpull.v1.XmlPullParserException;

/**
 * This class is responsible for constructing and sending the requests for the Web Service
 * Provider chosen by the user. Furthermore it processes the response from the WSP and creates
 * a list of Services based on the response.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class WspRequest extends LibertyMessage {

    private HttpTransport session;
    private Service[] services;

    /**
     * The constructor creates a WspRequest object based on the ResourceOffering.
     * 
     * @param offering ResourceOffering The ResouceOffering containing the url and token for
     *        the Web Service Provider.
     */
    public WspRequest(ResourceOffering offering) {
        super();
        this.offering = offering;
    }

    /**
     * This method creates the transport that will be used in the communication with the Web
     * Service Provider.
     * 
     * @return HttpTransport The transport that will be used by this class to send the query
     *         and receive the response.
     */
    public HttpTransport createTransport() {
        session = new HttpTransport(offering.getUrl());
        return session;
    }

    /**
     * This method constructs the request, sends it and processes the response.
     * 
     * @return Service[] A list of the Services returned from the WSP, containing service
     *         names, labels and values.
     * @throws MessageException If something wrong happens, an exception with the appropriate
     *         error code is thrown.
     */
    public Service[] sendRequest() throws MessageException {
        /* If the transport has not been created yet, create it: */
        if (session == null) {
            session = new HttpTransport(offering.getUrl());
        }
        /* Create the query itself first and the headers second to keep the timestamp as fresh as possible: */
        createBody();

        /* Creating the correlation header last to keep the timestamp fresh: */
        createSecurityHeader();
        createCorrelationHeader(false);
        headerOut = new Element[] { correlationHeader, securityHeader };

        try {
            /* Try to send the query: */
            session.call(offering.getSoapAction(), this);
        } catch (IOException e) {
            /* The communication was obstructed */
            throw new MessageException(Language.ERROR_WSP_LOSTCONTACT);
        } catch (XmlPullParserException e) {
            /* The response could not be parsed as XML */
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }

        try {
            /* Try to validate the correlation header: */
            processCorrelationHeader();
        } catch (MessageException me) {
            if (me.getErrorCode() == -1) {
                /* The timestamp of the correlation header deviated too much from the current time: */
                throw new MessageException(Language.ERROR_WSP_TIMESTAMPDEVIATION);
            } else {
                /* Some other exception occured: */
                throw me;
            }
        }

        try {
            /* The correlation header was ok, process the body: */
            processResponseBody("ServiceList");
        } catch (MessageException me) {
            if (me.getErrorCode() == Language.ERROR_INVALID_MESSAGE) {
                /* The message was invalid, maybe the WSP sent a fault instead: */
                processResponseBody("MobileRegisterFault");
            } else {
                /* Some other exception occured: */
                throw me;
            }
        }
        /* Success! Return the list of services */
        return services;
    }

    /**
     * Creates the request to be sent to the Web Service Provider.
     */
    private void createBody() {
        Element infoRequest = new Element();
        infoRequest.setPrefix("mms", Configurations.REQUEST_SCHEMA_URI);
        infoRequest.setNamespace(Configurations.REQUEST_SCHEMA_URI);
        infoRequest.setName("InformationRequest");
        infoRequest.addChild(Element.ELEMENT, offering.getResourceId());

        Element language = infoRequest.createElement(Configurations.REQUEST_SCHEMA_URI, "language");
        language.addChild(Element.TEXT, Configurations.LANGUAGE);
        infoRequest.addChild(Element.ELEMENT, language);
        bodyOut = infoRequest;
    }

    /**
     * This method processes the current element and stores the Services so that they can be
     * returned to the LibertyProtocol
     * 
     * @param current Element The element that is to be processed.
     * @throws MessageException If there are no results or the message is invalid.
     */
    protected void processResponse(Element current) throws MessageException {
        if (current.getNamespace().equals(Configurations.RESPONSE_SCHEMA_URI)) {
            /* This is a regular response */
            Element child = null;
            String name;
            String[][] values;
            Vector valueList;

            /* Create a dynamic list in case not all services are valid: */
            Vector list = new Vector(current.getChildCount());

            for (int i = 0; i < current.getChildCount(); i++) {
                child = current.getElement(i);

                if (child != null) {
                    if (child.getName().endsWith("Service")) {
                        /* Get the service name: */
                        name = child.getAttributeValue(child.getNamespace(), "name");
                        valueList = new Vector(child.getChildCount());
                        Element valueElement = null;
                        String[] value;

                        /* Go through the valueElements to get the service values: */
                        for (int j = 0; j < child.getChildCount(); j++) {
                            value = new String[2];
                            valueElement = child.getElement(j);

                            if (valueElement != null) {
                                /* Get the label: */
                                value[0] = valueElement.getAttributeValue(0);
                                /* Get the value: */
                                value[1] = valueElement.getAttributeValue(1);
                                valueList.addElement(value);
                            }
                        }

                        if (valueList.size() > 0) {
                            /* This service is ok - add it to the list. */
                            values = new String[valueList.size()][2];

                            for (int j = 0; j < valueList.size(); j++) {
                                values[j] = (String[]) valueList.elementAt(j);
                            }
                            list.addElement(new Service(name, values));
                        }
                    } else {
                        /* The response contains some unrecognised elements. */
                        throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                    }
                }
            }
            if (list.size() > 0) {
                /* The result is ok */
                services = new Service[list.size()];

                for (int i = 0; i < list.size(); i++) {
                    services[i] = (Service) list.elementAt(i);
                }
            } else {
                /* There were no results */
                throw new MessageException(Language.ERROR_WSP_NORESULTS);
            }
        } else if (current.getNamespace().equals(Configurations.FAULT_SCHEMA_URI)) {
            /* This response contains a fault message from the WSP */
            Element description = current.getElement(Configurations.FAULT_SCHEMA_URI,
                    "errorDescription");

            if (description != null) {
                String errorMsg = description.getText(0);
                if (errorMsg != null) {
                    throw new MessageException(Language.ERROR_WSP_RECEIVEDFAULT, errorMsg);
                } else {
                    throw new MessageException(Language.ERROR_INVALID_MESSAGE);
                }
            } else {
                throw new MessageException(Language.ERROR_INVALID_MESSAGE);
            }

        } else {
            throw new MessageException(Language.ERROR_INVALID_MESSAGE);
        }
    }
}
