/*
 * Classname: Controller
 * Version: 1.25
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.controller;

import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;

import mymobilesite.communication.*;

/**
 * The Controller class is the application's main class and controls the program execution. It
 * calls the other classes and handles all user and program (exceptions) events.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class Controller extends MIDlet implements LibertyProtocolListener {

    private Service[] serviceList;
    private String[] providerList;
    private GUI gui;
    private LibertyProtocol libProtocol;
    private String username;
    private MessageException me;
    private int chosenprovider;
    private boolean retrivedProviders;
    private String idpName;

    /**
     * The Controller constructor.
     */
    public Controller() {
        retrivedProviders = false;
    }

    /**
     * @return String The name of the identity provider as retrieved from the certificate.
     */
    public String getIdpName() {
        return idpName;
    }

    /**
     * @return Service[] An array of retrieved services.
     */
    public Service[] getServices() {
        return serviceList;
    }

    /**
     * @return String[] An array with the names of the service providers.
     */
    public String[] getProviderList() {
        return providerList;
    }

    /**
     * This method is called by the GUI when the user has pressed the button to start login.
     * The controller then calls the gui to change the window and tells the LibertyProtocol to
     * start login at the identity provider by passing on credentials.
     * 
     * @param un String The username which was entered into the login screen.
     * @param pwd String The password which was entered into the login screen.
     */
    public void startLogin(String un, String pwd) {
        gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_VERIFYING_PWD));
        /* store username so we can update Configurations: */
        username = un;
        /* start retrieval of providers: */
        libProtocol.getProviders(un, pwd);
    }

    /**
     * This method request the LibertyProtocol to start retrieving services from the chosen
     * provider. It also changes gui to a waiting window.
     * 
     * @param index int The index in the list of providers which the user has chosen.
     */
    public void contactWsp(int index) {
        chosenprovider = index;
        libProtocol.requestService(index);
        gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_GETTING_SERVICES));
    }

    /**
     * This method is used by the GUI to alert the controller that the user pressed back while
     * waiting.
     * 
     * @param waitingState int The waiting window which the back command was called from.
     */
    public void userAborted() {
        libProtocol.abort();
        switch (gui.getPreviousState()) {
        case GUI.LOGIN_SCREEN:
            gui.setWindow(GUI.LOGIN_SCREEN, idpName);
            break;
        case GUI.PROVIDER_SCREEN:
            gui.setWindow(GUI.PROVIDER_SCREEN);
            break;
        }
    }

    /**
     * This method is called by the GUI to show that the user has requested the application to
     * try again the previous action which failed.
     */
    public void retry() {
        if (((me.getErrorCode()) == Language.ERROR_IDP_LOSTCONTACT)
                || ((me.getErrorCode()) == Language.ERROR_DS_LOSTCONTACT)
                || ((me.getErrorCode()) == Language.ERROR_IDP_NOTTRUSTED)
                || ((me.getErrorCode()) == Language.ERROR_IDP_NOTFOUND)
                || ((me.getErrorCode()) == Language.ERROR_IDPNAME_NOTFOUND)
                || ((me.getErrorCode()) == Language.ERROR_REPLAY_DETECTED)) {
            gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_IDP_SESSION));
            libProtocol.establishIdpSession();
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_AUTH_NOTACCEPTED
                || ((me.getErrorCode()) == Language.ERROR_DS_NORESULTS)
                || ((me.getErrorCode()) == Language.ERROR_IDP_FAULT)) {
            gui.setWindow(GUI.LOGIN_SCREEN, idpName);
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_WSP_LOSTCONTACT) {
            if (providerList.length > 1) {
                gui.setWindow(GUI.PROVIDER_SCREEN);
            } else {
                gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_GETTING_SERVICES));
                libProtocol.requestService(chosenprovider);
            }
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_WSP_NORESULTS) {
            gui.setWindow(GUI.PROVIDER_SCREEN);
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_WSP_RECEIVEDFAULT) {
            gui.setWindow(GUI.PROVIDER_SCREEN);
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_IDP_TIMESTAMPDEVIATION) {
            gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_VERIFYING_PWD));
            libProtocol.retryGetProviders();
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_DS_TIMESTAMPDEVIATION) {
            gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_FINDING_WSP));
            libProtocol.retryDiscoQuery();
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_WSP_TIMESTAMPDEVIATION) {
            gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_GETTING_SERVICES));
            libProtocol.requestService(chosenprovider);
            me = null;
        } else if ((me.getErrorCode()) == Language.ERROR_INVALID_MESSAGE) {
            if (!retrivedProviders) {
                gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_IDP_SESSION));
                libProtocol.establishIdpSession();
            } else {
                gui.setWindow(GUI.PROVIDER_SCREEN);
            }
            me = null;
        }
    }

    /**
     * Starts the application.
     * 
     * @see javax.microedition.midlet.MIDlet#startApp()
     */
    protected void startApp() throws MIDletStateChangeException {
        gui = new GUI(this);
        gui.setWindow(GUI.SPLASH_SCREEN);
        (new Configurations()).read();
        gui.setLanguage(Configurations.LANGUAGE);
        libProtocol = new LibertyProtocol(this);
        libProtocol.establishIdpSession();
    }

    /**
     * This application does not support pausing the application and calls destroyApp.
     * 
     * @see javax.microedition.midlet.MIDlet#pauseApp()
     */
    protected void pauseApp() {
        try {
            destroyApp(true);
        } catch (Exception e) {
        }
    }

    /**
     * @see javax.microedition.midlet.MIDlet#destroyApp(boolean)
     */
    protected void destroyApp(boolean unconditional) throws MIDletStateChangeException {
        this.notifyDestroyed();
    }

    /**
     * Receives the results from the LibertyProtocol and handles them according to the state.
     * 
     * @see mymobilesite.controller.LibertyProtocolListener#resultsReady(int, java.lang.Object)
     */
    public void resultsReady(int state, Object result) {
        switch (state) {
        case LibertyProtocol.CONTACTING_IDP:
            idpName = (String) result;
            gui.setWindow(GUI.LOGIN_SCREEN, idpName);
            break;
        case LibertyProtocol.AUTHENTICATING_USER:
            gui.setWindow(GUI.WAITING_SCREEN, Language.getWaitingMsg(Language.WAITING_FINDING_WSP));
            /* Check if the username used in this authentication is the same as in the RecordStore: */
            if (!username.equals(Configurations.USERNAME)) {
                (new Configurations()).write(Configurations.USERNAME_INDEX, username);
            }
            break;
        case LibertyProtocol.RETRIEVING_PROVIDER_LIST:
            retrivedProviders = true;
            providerList = (String[]) result;
            gui.setWindow(GUI.PROVIDER_SCREEN);
            break;
        case LibertyProtocol.REQUESTING_SERVICE:
            serviceList = (Service[]) result;
            gui.setWindow(GUI.SERVICELIST_SCREEN);
            break;
        }
    }

    /**
     * Calls the GUI to display an appropriate error message.
     * 
     * @see mymobilesite.controller.LibertyProtocolListener#encounteredException(int,
     *      mymobilesite.communication.MessageException)
     */
    public void encounteredException(int currentState, MessageException me) {
        this.me = me;
        gui.setWindow(GUI.ERROR_SCREEN, me.getMessage());
    }
}
