/*
 * Classname: GUI
 * Version: 1.49
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package mymobilesite.controller;

import java.io.IOException;

import javax.microedition.lcdui.Alert;
import javax.microedition.lcdui.AlertType;
import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Choice;
import javax.microedition.lcdui.ChoiceGroup;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.CommandListener;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Displayable;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;
import javax.microedition.lcdui.ImageItem;
import javax.microedition.lcdui.Item;
import javax.microedition.lcdui.ItemCommandListener;
import javax.microedition.lcdui.List;
import javax.microedition.lcdui.StringItem;
import javax.microedition.lcdui.TextField;
import javax.microedition.lcdui.game.GameCanvas;

/**
 * The GUI class does everything that is presented to the user and handles events that are
 * caused by user action. The class relies on several states which indicates which GUI window
 * should be shown to the user.
 * 
 * @author Gunn Olaussen, Kirsti N. Torgersen
 */
public class GUI implements CommandListener, ItemCommandListener, Runnable {

    /** These state variables are used by the class to determine what window to show. */
    public static final int SPLASH_SCREEN = 0;
    public static final int LOGIN_SCREEN = 1;
    public static final int HELP_SCREEN = 2;
    public static final int SETTINGS_SCREEN = 3;
    public static final int PROVIDER_SCREEN = 4;
    public static final int SERVICELIST_SCREEN = 5;
    public static final int SERVICEDETAIL_SCREEN = 6;
    public static final int WAITING_SCREEN = 7;
    public static final int ERROR_SCREEN = 8;

    private Controller controller;
    private int currentState;
    private int previousState;
    private int userChoice;
    private Language language;
    private Service[] serviceList;
    private String message;

    private Display display;
    private Form mainForm;
    private Command ok;
    private Command exit;
    private Command settings;
    private Command help;
    private Command save;
    private Command back;
    private Command tryagain;
    private TextField username;
    private TextField password;
    private ChoiceGroup lang;
    private ChoiceGroup prov;
    private ChoiceGroup servicel;

    /**
     * The constructor is given a reference to the controller so that it can retrieve
     * neccessary variables. All the GUI variables which are needed throughout the
     * application's execution are initialized here.
     * 
     * @param cont Controller The program controller which is needed to retrieve some
     *        variables.
     */
    public GUI(Controller cont) {
        controller = cont;
        display = Display.getDisplay(cont);
    }

    /**
     * This returns the privious state which is used when the user is going back to the
     * previous screen.
     * 
     * @return int The previous application state
     */
    public int getPreviousState() {
        return previousState;
    }

    /**
     * This method tells the application which window it should show. This is used when there
     * is a window which does not show a message which is expected as an input.
     * 
     * @param window int The id of the window which we want displayed
     */
    public void setWindow(int window) {
        setWindow(window, null);
    }

    /**
     * This method tells the application which window it should show. This is used when the
     * user should be shown a specific message.
     * 
     * @param window int The id of the window which we want displayed
     * @param msg String The message that should be shown in this window
     */
    public void setWindow(int window, String msg) {
        currentState = window;
        message = msg;
        (new Thread(this)).start();
    }

    /**
     * This method sets the language in the Configuration class if the input is different from
     * that currently in the RecordStore. The language used in the application is also updated.
     * 
     * @param lan String The chosen language.
     */
    public void setLanguage(String lan) {
        Configurations.LANGUAGE = lan;
        language = Language.getLanguage();
        createCommandButtons();
    }

    /**
     * Creates new command buttons either at startup or when the language is changed.
     */
    private void createCommandButtons() {
        ok = new Command(language.command_ok, Command.OK, 1);
        exit = new Command(language.command_exit, Command.OK, 2);
        settings = new Command(language.command_settings, Command.OK, 3);
        help = new Command(language.command_help, Command.OK, 4);
        save = new Command(language.command_save, Command.OK, 5);
        back = new Command(language.command_back, Command.OK, 6);
        tryagain = new Command(language.command_tryagain, Command.OK, 7);
    }

    /**
     * This method is used when a window is set. It runs in its own Thread and creates the
     * window which is determined by the current state.
     * 
     * @see java.lang.Runnable#run()
     */
    public void run() {
        switch (currentState) {
        case SPLASH_SCREEN:
             SplashScreen splash = new SplashScreen();
             display.setCurrent(splash);
            break;
        case WAITING_SCREEN:
            mainForm = new Form(language.label_applName);
            mainForm.append(new StringItem(message, "\n\n" + language.label_waiting));

            /* the first window of the application should not have the "back" option: */
            if (!message.equals(Language.getWaitingMsg(Language.WAITING_IDP_SESSION))){
                mainForm.addCommand(back);
            }

            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        case LOGIN_SCREEN:
            mainForm = new Form(language.label_applName);

            StringItem idpname = new StringItem(language.label_idp + "\n", message, StringItem.HYPERLINK);
            idpname.addCommand(ok);
            idpname.addCommand(exit);
            idpname.addCommand(settings);
            idpname.addCommand(help);
            idpname.setItemCommandListener(this);

            username = new TextField(language.label_un, Configurations.USERNAME, 11, TextField.ANY);
            username.addCommand(ok);
            username.addCommand(exit);
            username.addCommand(settings);
            username.addCommand(help);
            username.setItemCommandListener(this);

            password = new TextField(language.label_pwd, null, 11, TextField.PASSWORD);
            password.addCommand(ok);
            password.addCommand(exit);
            password.addCommand(settings);
            password.addCommand(help);
            password.setItemCommandListener(this);

            mainForm.append(idpname);
            mainForm.append(username);
            mainForm.append(password);

            display.setCurrent(mainForm);
            previousState = LOGIN_SCREEN;
            break;
        case HELP_SCREEN:
            mainForm = new Form(language.label_applName);
            mainForm.append(new StringItem(language.label_help, "\n" + language.help));
            mainForm.addCommand(back);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        case SETTINGS_SCREEN:
            mainForm = new Form(language.label_applName);
            lang = new ChoiceGroup(language.label_language, ChoiceGroup.EXCLUSIVE);
            int index;

            for (int i = 0; i < (language.languageList).length; i++) {
                /* add each supported language to the list of choices: */
                lang.append(language.languageList[i], null);

                /* retrives the language code and sets the default language as selected: */
                index = (language.languageList[i]).indexOf("(");
                if (((language.languageList[i]).substring(index + 1, index + 3))
                        .equals(Configurations.LANGUAGE)) {
                    lang.setSelectedIndex(i, true);
                }
            }
            mainForm.append(lang);
            mainForm.addCommand(save);
            mainForm.addCommand(back);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        case PROVIDER_SCREEN:
            /* removing old object references which will no longer be used: */
            username = null;
            password = null;
            lang = null;
            settings = null;
            help = null;
            
            mainForm = new Form(language.label_applName);
            /* making a list to hold provider names: */
            prov = new ChoiceGroup(language.label_providers, Choice.EXCLUSIVE);
            String[] list = controller.getProviderList();
            
            for (int i = 0; i < list.length; i++) {
                /* add each available provider to the list of choices: */
                prov.append(list[i], null);
            }
            
            prov.setSelectedIndex(0, true);
            mainForm.append(prov);
            mainForm.addCommand(ok);
            mainForm.addCommand(exit);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            previousState = PROVIDER_SCREEN;
            break;
        case SERVICELIST_SCREEN:
            /* removing an old object reference: */
            prov = null;

            mainForm = new Form(language.label_applName);
            servicel = new ChoiceGroup(language.label_services, ChoiceGroup.EXCLUSIVE);
            serviceList = controller.getServices();
            
            for (int i = 0; i < serviceList.length; i++) {
                /* add each available service to the list of choices: */
                servicel.append(serviceList[i].getName(), null);
            }
            servicel.setSelectedIndex(0, true);
            mainForm.append(servicel);
            mainForm.addCommand(ok);
            mainForm.addCommand(exit);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        case SERVICEDETAIL_SCREEN:
            mainForm = new Form(language.label_applName);
            Service s = serviceList[userChoice];
            mainForm.append(new StringItem(s.getName(), ":\n"));
            String[][] values = s.getValue();

            for (int i = 0; i < values.length; i++) {
                /* adding each of the label/value pairs: */
                mainForm.append(new StringItem(values[i][0] + ":\n", values[i][1]));
            }
            mainForm.addCommand(ok);
            mainForm.addCommand(exit);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        case ERROR_SCREEN:
            mainForm = new Form(language.label_applName);
            mainForm.append(new StringItem(language.label_error, "\n\n" + message));
            mainForm.addCommand(tryagain);
            mainForm.addCommand(exit);
            mainForm.setCommandListener(this);
            display.setCurrent(mainForm);
            break;
        }
    }

    /**
     * Calls doCommand(c)
     * 
     * @see javax.microedition.lcdui.CommandListener#commandAction(javax.microedition.lcdui.Command,
     *      javax.microedition.lcdui.Displayable)
     */
    public void commandAction(Command c, Displayable d) {
        doCommand(c);
    }

    /**
     * Calls doCommand(c)
     * 
     * @see javax.microedition.lcdui.ItemCommandListener#commandAction(javax.microedition.lcdui.Command,
     *      javax.microedition.lcdui.Item)
     */
    public void commandAction(Command c, Item item) {
        doCommand(c);
    }

    /**
     * Performs the appropriate action according to the command it receives.
     */
    private void doCommand(Command c) {
        if (c.equals(exit)) {
            controller.notifyDestroyed();
        } else if (c.equals(ok)) {
            switch (currentState) {
            case LOGIN_SCREEN:
                if (((username.getString()).length() > 0)
                        && ((password.getString()).length() > 0)) {
                    controller.startLogin(username.getString(), password.getString());
                } else {
                    /* Do nothing if the username or password have no value yet */
                }
                break;
            case PROVIDER_SCREEN:
                controller.contactWsp(prov.getSelectedIndex());
                break;
            case SERVICELIST_SCREEN:
                userChoice = servicel.getSelectedIndex();
                setWindow(SERVICEDETAIL_SCREEN);
                break;
            case SERVICEDETAIL_SCREEN:
                setWindow(SERVICELIST_SCREEN);
                break;
            }
        } else if (c.equals(settings)) {
            setWindow(SETTINGS_SCREEN);
        } else if (c.equals(help)) {
            setWindow(HELP_SCREEN);
        } else if (c.equals(save)) {
            /* find which language was chosen and update Configurations: */
            int choice = lang.getSelectedIndex();
            int index = (language.languageList[choice]).indexOf("(");
            String languageString = (language.languageList[choice]).substring(index + 1,
                    index + 3);
            if (!languageString.equals(Configurations.LANGUAGE)) {
                setLanguage(languageString);
            }
            setWindow(LOGIN_SCREEN, controller.getIdpName());
        } else if (c.equals(back)) {
            switch (currentState) {
            case SETTINGS_SCREEN:
                setWindow(LOGIN_SCREEN, controller.getIdpName());
                break;
            case HELP_SCREEN:
                setWindow(LOGIN_SCREEN, controller.getIdpName());
                break;
            case WAITING_SCREEN:
                controller.userAborted();
                break;
            }
        } else if (c.equals(tryagain)) {
            controller.retry();
        }
    }
}

class SplashScreen extends Canvas {

    private Image image;
    
    /**
     * The constructor loads the image that will be shown in the splash screen
     */
    public SplashScreen() {
        super();
        try {
            image = Image.createImage(getClass().getResourceAsStream("mymobilesite.png"));
        } catch (IOException e) {}
    }

    /** 
     * @see javax.microedition.lcdui.Canvas#paint(javax.microedition.lcdui.Graphics)
     */
    protected void paint(Graphics g) {
        this.setFullScreenMode(true);
        g.setColor(10, 10, 10);
        g.fillRect(0, 0, this.getWidth(), this.getHeight());
        if (image != null) {
            int imageX = (int)((this.getWidth()-image.getWidth())/2);
            int imageY = (int)((this.getHeight()-image.getHeight())/2);
            g.drawImage(image, imageX, imageY,Graphics.TOP|Graphics.LEFT);
        } else {
            String mms = "MyMobileSite";
            Font f = g.getFont();
            int textX = (int)((this.getWidth()-f.stringWidth(mms))/2);
            g.setColor(0, 0, 0);
            g.drawString(mms, textX, 0, Graphics.TOP|Graphics.LEFT);
        }
    }
}
