/*
 * Classname: Register
 * Version: 1.10
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package loanfund;

import java.io.*;
import javax.xml.parsers.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * The Register class is used to access the services which are stored in an XML file. It only
 * has one method which finds the service values for the current user.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class Register {
    private final String dbFile = "wsp_db.xml";
    private String[] values;
    private String id;

    /**
     * The constructor stores the username which should be used when looking for services.
     * 
     * @param idnr String The username/id for the user identified in the service request
     */
    public Register(String idnr) {
        id = idnr;
    }

    /**
     * This method reads the services from the XML file and returns the values for the user
     * identified in the id variable.
     * 
     * @return String[] The found services for the user
     */
    public String[] readValues() {
        values = new String[9];

        /* reading xml file: */
        Document doc;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            doc = builder.parse(dbFile);
            Element root = doc.getDocumentElement();
            NodeList profiles = root.getElementsByTagName("profile");

            Element currentprofile;
            String currentid;

            /* checks through all the profiles: */
            for (int i = 0; i < profiles.getLength(); i++) {
                currentprofile = (Element) profiles.item(i);
                currentid = currentprofile.getElementsByTagName(
                        "id").item(0).getChildNodes().item(0).getNodeValue();

                if (currentid.equals(id)) {
                    values[0] = getDebt(currentprofile);

                    if (hasLastInstalmentTag(currentprofile)) {
                        NodeList lastinstalment = currentprofile.getElementsByTagName(
                                "lastinstalment").item(0).getChildNodes();
                        values[1] = getInstalmentDate(lastinstalment);
                        values[2] = getInstalmentSum(lastinstalment);
                    }

                    if (hasNextInstalmentTag(currentprofile)) {
                        NodeList lastinstalment = currentprofile.getElementsByTagName(
                                "nextinstalment").item(0).getChildNodes();
                        values[3] = getInstalmentDate(lastinstalment);
                        values[4] = getInstalmentSum(lastinstalment);
                    }

                    if (hasApplicationTag(currentprofile)) {
                        NodeList application = currentprofile.getElementsByTagName(
                                "application").item(0).getChildNodes();
                        values[5] = getApplicationReceivedDate(application);
                        values[6] = getApplicationStatus(application);

                        if (hasApplicationSumTag(application)) {
                            NodeList sum = application.item(5).getChildNodes();
                            values[7] = getApplicationSumGrant(sum);
                            values[8] = getApplicationSumLoan(sum);
                        }
                    }
                    return values;
                }
            }
        } catch (ParserConfigurationException pce) {
            /* if an error occurs null is return: */
            values = null;
            if (RegisterService.DEBUG) Log.write(pce);
        } catch (IOException ioe) {
            /* if an error occurs null is return: */
            values = null;
            if (RegisterService.DEBUG) Log.write(ioe);
        } catch (SAXException sax) {
            /* if an error occurs null is return: */
            values = null;
            if (RegisterService.DEBUG) Log.write(sax);
        }

        return null;
    }

    /**
     * Used to check if the sum tag inside the application Node exists.
     */
    private boolean hasApplicationSumTag(NodeList application) {
        for (int i = 0; i < application.getLength(); i++) {
            if (application.item(i).getNodeName().equals("sum")) {
                return true;
            }
        }
        return false;
    }

    /**
     * Used to check if this user profile has an application tag
     */
    private boolean hasApplicationTag(Element currentprofile) {
        return currentprofile.getElementsByTagName("application").getLength() > 0;
    }

    /**
     * Used to check id this user profile has a nextinstalment tag
     */
    private boolean hasNextInstalmentTag(Element currentprofile) {
        return currentprofile.getElementsByTagName("nextinstalment").getLength() > 0;
    }

    /**
     * Used to check id this user profile has a lastinstalment tag
     */
    private boolean hasLastInstalmentTag(Element currentprofile) {
        return currentprofile.getElementsByTagName("lastinstalment").getLength() > 0;
    }

    /**
     * Finds the application date value from the NodeList
     */
    private String getApplicationReceivedDate(NodeList application) {
        return application.item(1).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the application status value from the NodeList
     */
    private String getApplicationStatus(NodeList application) {
        return application.item(3).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the application grant sum value from the NodeList
     */
    private String getApplicationSumGrant(NodeList sum) {
        return sum.item(1).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the application loan sum value from the NodeList
     */
    private String getApplicationSumLoan(NodeList sum) {
        return sum.item(3).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the instalment sum value from the NodeList
     */
    private String getInstalmentSum(NodeList instalment) {
        return instalment.item(3).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the instalment date value from the NodeList
     */
    private String getInstalmentDate(NodeList instalment) {
        return instalment.item(1).getChildNodes().item(0).getNodeValue();
    }

    /**
     * Finds the debt value from the NodeList
     */
    private String getDebt(Element currentprofile) {
        return currentprofile.getElementsByTagName("debt").item(0).getTextContent();
    }
}
