/*
 * Classname: RegisterFault
 * Version: 1.9
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package loanfund;

import java.util.Date;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;
import javax.xml.soap.SOAPMessage;

import org.kobjects.isodate.IsoDate;

/**
 * This method is used when a ServiceResponse cannot be created. This message sends an error to
 * alert the mobile application of what went wrong.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class RegisterFault {

    /* The error codes: */
    public static final String UNKNOWN_ID = "unknownId";
    public static final String TIMED_OUT = "requestTimedOut";
    public static final String SERVICE_UNAVAILABLE = "serviceUnavailable";
    public static final String NOT_AUTHORIZED = "notAuthorized";
    public static final String INTERNAL_ERROR = "internalError";

    /* The English messages related to the error codes: */
    private static final String UNKNOWN_ID_MESSAGE_GB = 
        "The username is unknown to this service.";
    private static final String TIMED_OUT_MESSAGE_GB = 
        "Time out before the service could create a response.";
    private static final String SERVICE_UNAVAILABLE_MESSAGE_GB = 
        "The service is temporarily unavailable.";
    private static final String NOT_AUTHORIZED_MESSAGE_GB = 
        "Access not authorized due to faulty info from log-in.";
    private static final String INTERNAL_ERROR_MESSAGE_GB = 
        "An internal error occured in the Educational Loan Fund system.";

    /* The Norwegian messages related to the error codes: */
    private static final String UNKNOWN_ID_MESSAGE_NB = 
        "Brukernavnet er ukjent for denne tjenesten.";
    private static final String TIMED_OUT_MESSAGE_NB = 
        "Tidsutkobling fr tjenesten kunne lage et svar.";
    private static final String SERVICE_UNAVAILABLE_MESSAGE_NB = 
        "Tjenesten er for yeblikket utilgjengelig.";
    private static final String NOT_AUTHORIZED_MESSAGE_NB = 
        "Tilgang ikke godkjent grunnet feil info fra innlogging.";
    private static final String INTERNAL_ERROR_MESSAGE_NB = 
        "Det skjedde en feil internt i Lnekassens systemer.";

    private static final String ns = 
        "http://fedoralab1.kantega.no:8080/faults/mymobilesite-faults.xsd";
    private static final String sb = 
        "urn:liberty:sb:2003-08";

    private ServiceRequestException fault;
    private String lang;
    private String inResponseTo;
    private SOAPMessage message;
    private SOAPEnvelope env;

    /**
     * The constructor receives the Exception that occured, an inResponseTo String which is
     * used to relate the error to something and the language to use. The Exception is the
     * reason why a fault message is created instead of the ordinary response.
     * 
     * @param fault ServiceRequestException The exeception that occured
     * @param inResponseTo String The messageId of the request
     * @param lang String The langugage we want the fault to use for the error
     */
    public RegisterFault(ServiceRequestException fault, String inResponseTo, String lang) {
        this.fault = fault;
        this.inResponseTo = inResponseTo;
        this.lang = lang;
        try {
            message = MessageFactory.newInstance().createMessage();
            message.setProperty(SOAPMessage.CHARACTER_SET_ENCODING, "UTF-8");
            message.setProperty(SOAPMessage.WRITE_XML_DECLARATION, "true");

            env = message.getSOAPPart().getEnvelope();
            createHeader();
            createBody();
        } catch (Exception e) {
            if (RegisterService.DEBUG) Log.write(e);
        }
    }

    /**
     * This creates the message header in which it tries to set the correlation, timestamp and
     * messageID. Otherwise an SOAPException is thrown.
     */
    private void createHeader() throws SOAPException {
        SOAPHeader header = message.getSOAPHeader();
        SOAPHeaderElement correlation = header.addHeaderElement(env.createName("Correlation",
                "sb", sb));
        correlation.addAttribute(env.createName("messageID", "sb", "urn:liberty:sb:2003-08"),
                Randomizer.nextID());
        if (inResponseTo != null){
            correlation.addAttribute(env.createName("refToMessageID", "sb", sb), inResponseTo);
        }
        correlation.addAttribute(env.createName("timestamp", "sb", sb), IsoDate.dateToString(
                new Date(System.currentTimeMillis()), IsoDate.DATE_TIME));
        correlation.setMustUnderstand(true);
    }

    /**
     * This method creates the message body which contains an error code and description.
     */
    private void createBody() throws SOAPException {
        String errorCode = fault.getMessage();

        SOAPBody body = message.getSOAPBody();
        SOAPElement mobileRegisterFault = body.addBodyElement(env.createName(
                "MobileRegisterFault", "ns", ns));

        SOAPElement errorCodeElement = mobileRegisterFault.addChildElement(env.createName(
                "errorCode", "ns", ns));
        errorCodeElement.addTextNode(errorCode);

        SOAPElement errorMessageElement = mobileRegisterFault.addChildElement(env.createName(
                "errorDescription", "ns", ns));

        String errorMessage = "no message";

        if (errorCode.equals(UNKNOWN_ID)) {
            errorMessage = (lang.equals("nb") ? UNKNOWN_ID_MESSAGE_NB : UNKNOWN_ID_MESSAGE_GB);
        } else if (errorCode.equals(TIMED_OUT)) {
            errorMessage = (lang.equals("nb") ? TIMED_OUT_MESSAGE_NB : TIMED_OUT_MESSAGE_GB);
        } else if (errorCode.equals(SERVICE_UNAVAILABLE)) {
            errorMessage = (lang.equals("nb") ? SERVICE_UNAVAILABLE_MESSAGE_NB
                    : SERVICE_UNAVAILABLE_MESSAGE_GB);
        } else if (errorCode.equals(NOT_AUTHORIZED)) {
            errorMessage = (lang.equals("nb") ? NOT_AUTHORIZED_MESSAGE_NB
                    : NOT_AUTHORIZED_MESSAGE_GB);
        } else if (errorCode.equals(INTERNAL_ERROR)) {
            errorMessage = (lang.equals("nb") ? INTERNAL_ERROR_MESSAGE_NB
                    : INTERNAL_ERROR_MESSAGE_GB);
        }
        errorMessageElement.addTextNode(errorMessage);
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return message.toString();
    }
}
