/*
 * Classname: RegisterService
 * Version: 1.14
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package loanfund;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.servlet.ServletInputStream;
import javax.servlet.http.HttpServletRequest;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.MimeHeaders;
import javax.xml.soap.SOAPMessage;

/**
 * The RegisterService is the controller class in the web service provider. It starts the
 * request parsing and response creation. If it detects a ServiceRequestException it creats a
 * RegisterFault message instead.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class RegisterService {

    public static final String LANGUAGECODE_NB = "nb";
    public static final String LANGUAGECODE_GB = "gb";

    public static boolean DEBUG = true;

    /**
     * This method is the static method called from the JSP page to make a request for the web
     * service provider. This method parses the request and creates a response if possible. If
     * the query is properly formated and the data requested is present, a response containing
     * the service offerings is made. Otherwise a fault message is returned.
     * 
     * @param req HttpServletRequest The web service provider request
     * @return String The response message
     */
    public static String getService(HttpServletRequest req) {
        ServletInputStream sis = null;
        String s = "";
        ServiceRequest sreq = null;
        try {
            sis = req.getInputStream();
            MessageFactory msgFact = MessageFactory.newInstance();
            SOAPMessage doc = msgFact.createMessage(new MimeHeaders(), sis);
            sreq = new ServiceRequest(doc);
            ServiceResponse sres = new ServiceResponse(sreq.getMessageID(), sreq.getLanguage());
            Register r = new Register(sreq.getUserID());
            String[] result = r.readValues();

            if (result != null) {
                sres.createBody(result);
                s = sres.toString();
                if (sreq.sessionTimedOut())
                    throw new ServiceRequestException(RegisterFault.TIMED_OUT);
                if (DEBUG) debug(sreq, sres);
            } else {
                throw new ServiceRequestException(RegisterFault.UNKNOWN_ID);
            }
        } catch (ServiceRequestException e) {
            s = createFault(sreq, e);
        } catch (Exception e) {
            if (DEBUG) Log.write(e);
            s = createFault(sreq, new ServiceRequestException(RegisterFault.INTERNAL_ERROR));
        } finally {
            try {
                if (sis != null) sis.close();
            } catch (IOException e) {
                if (DEBUG) Log.write(e);
            }
        }
        return s;
    }

    /**
     * This method create a new Fault message based on the ServiceRequestException.
     */
    private static String createFault(ServiceRequest sreq, ServiceRequestException e) {
        String s;
        if (RegisterService.DEBUG) Log.write(e);
        String lang = (sreq != null ? sreq.getLanguage() : "gb");
        String inResponseTo = (sreq != null ? sreq.getMessageID() : null);
        RegisterFault fault = new RegisterFault(e, inResponseTo, lang);
        s = fault.toString();
        if (DEBUG) debug(sreq, fault);
        return s;
    }

    /**
     * This method is used when in debug mode to write the last incomming request/response pair
     * to a file.
     * 
     * @param req ServiceRequest The incomming request
     * @param res ServiceResponse The response that was created
     */
    public static void debug(ServiceRequest req, ServiceResponse res) {
        try {
            File f = new File("/home/apache-tomcat-5.5.17/logs/WSPdebug.txt");
            if (!f.exists()) {
                f.createNewFile();
            }
            FileOutputStream fos = new FileOutputStream(f);
            String s = "MessageID: " + req.getMessageID() + "\n";
            fos.write("-----------Request: ----------\n".getBytes());
            fos.write(s.getBytes());
            s = "Language: " + req.getLanguage() + "\n";
            fos.write(s.getBytes());
            s = "UserID: " + req.getUserID() + "\n";
            fos.write(s.getBytes());
            s = "Timestamp: " + req.getTimestamp() + "\n";
            fos.write(s.getBytes());
            s = req.toString();
            fos.write(s.getBytes());
            fos.write("\n-----------Response: ----------\n".getBytes());
            s = res.toString();
            fos.write(s.getBytes());
            fos.flush();

        } catch (Exception e) {
            if (DEBUG) Log.write(e);
        }
    }

    /**
     * This method is used when in debug mode to write the last incomming request/response pair
     * to a file.
     * 
     * @param req ServiceRequest The incomming request
     * @param res RegisterFault The fault message that was created
     */
    public static void debug(ServiceRequest req, RegisterFault fault) {
        try {
            File f = new File("/home/apache-tomcat-5.5.17/logs/WSPdebug.txt");
            if (!f.exists()) {
                f.createNewFile();
            }
            FileOutputStream fos = new FileOutputStream(f);
            String s = null;
            if (req != null) {
                s = "MessageID: " + req.getMessageID() + "\n";
                fos.write("-----------Request: ----------\n".getBytes());
                fos.write(s.getBytes());
                s = "Language: " + req.getLanguage() + "\n";
                fos.write(s.getBytes());
                s = "UserID: " + req.getUserID() + "\n";
                fos.write(s.getBytes());
                s = "Timestamp: " + req.getTimestamp() + "\n";
                fos.write(s.getBytes());
                s = req.toString();
                fos.write(s.getBytes());
            }
            fos.write("\n-----------Fault: -------------\n".getBytes());
            s = fault.toString();
            fos.write(s.getBytes());
            fos.flush();

        } catch (Exception e) {
            if (DEBUG) Log.write(e);
        }
    }
}
