/*
 * Classname: ServiceRequest
 * Version: 1.13
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package loanfund;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.Signature;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.Iterator;

import javax.xml.soap.*;

import org.kobjects.base64.Base64;
import org.kobjects.isodate.IsoDate;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * The ServiceRequest is used to parse the request and save the information that we need to
 * verify the validity of the message and to find the information needed in the response.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class ServiceRequest {

    private static final long MAX_TIME_DEVIATION = 300000;
    private static final long SESSION_TIME_TO_LIVE = 600000;

    private Date sessiontimeout;
    private Date timestamp;
    private Node authorizationAssertion;
    private Node authenticationAssertion;
    private SOAPMessage message;
    private String authenticationIssuer;
    private String authorizationIssuer;
    private String language;
    private String messageid;
    private String resourceid;
    private String userid;

    /**
     * The constructor tries to parse the body and the header of the SOAP message. Throwing a
     * ServiceRequestException means that it is not possible to create a proper response.
     * 
     * @param message SOAPMessage The SOAPMessage containing the request
     * @throws ServiceRequestException Thrown to show that the request parsing was unsuccessful
     */
    public ServiceRequest(SOAPMessage message) throws ServiceRequestException {
        this.message = message;
        try {
            parseBody(message.getSOAPBody());
            parseHeader(message.getSOAPHeader());
        } catch (GeneralSecurityException e) {
            if (RegisterService.DEBUG) Log.write(e);
            throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
        } catch (SOAPException e) {
            if (RegisterService.DEBUG) Log.write(e);
            throw new ServiceRequestException(RegisterFault.INTERNAL_ERROR);
        }
    }

    /**
     * This method parses the header of the message and checks the timestamp for possible
     * replay of messages. The authorisation assertion is processed by a separate method.
     */
    private void parseHeader(SOAPHeader header) throws SOAPException, ServiceRequestException,
            GeneralSecurityException {
        Iterator it = header.extractAllHeaderElements();
        SOAPHeaderElement el;
        
        while (it.hasNext()) {
            el = (SOAPHeaderElement) it.next();
            if (el.getNodeName().contains("Correlation")) {
                Iterator atts = el.getAllAttributes();
                Name s;
                while (atts.hasNext()) {
                    s = ((Name) atts.next());
                    if (s.getQualifiedName().contains("messageID"))
                        messageid = el.getAttributeValue(s);
                    else if (s.getQualifiedName().contains("timestamp")) {
                        timestamp = IsoDate.stringToDate(el.getAttributeValue(s), IsoDate.DATE_TIME);
                        if (Math.abs(timestamp.getTime() - System.currentTimeMillis()) > MAX_TIME_DEVIATION) {
                            throw new ServiceRequestException(RegisterFault.TIMED_OUT);
                        }
                    }
                }
            } else if (el.getNodeName().contains("Security")) {
                processAuthorizationToken(el, 0);
            }
        }
    }

    /**
     * This method processes the authorisation assertion that was in the header.
     */
    private void processAuthorizationToken(Node token, int parsingDebth) throws SOAPException,
            ServiceRequestException, GeneralSecurityException {
        int nextDebth = parsingDebth + 1;
        if (token.hasChildNodes()) {
            NodeList list = token.getChildNodes();
            Node child;
            
            for (int i = 0; i < list.getLength(); i++) {
                child = list.item(i);
                if (child.getNodeName() == null) {
                    /* Do nothing */
                } else if (child.getNodeName().endsWith("Assertion")) {
                    authorizationAssertion = child.cloneNode(true);
                    NamedNodeMap atts = child.getAttributes();
                    
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("Issuer")) {
                            authorizationIssuer = atts.item(j).getTextContent();
                        }
                    }
                } else if (child.getNodeName().endsWith("Conditions")) {
                    NamedNodeMap atts = child.getAttributes();
                    
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("NotBefore")) {
                            Date notBefore = IsoDate.stringToDate(atts.item(j)
                                    .getTextContent(), IsoDate.DATE_TIME);
                            if (notBefore.after(new Date(System.currentTimeMillis()))) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        } else if (atts.item(j).getNodeName().endsWith("NotOnOrAfter")) {
                            Date notOnOrAfter = IsoDate.stringToDate(atts.item(j)
                                    .getTextContent(), IsoDate.DATE_TIME);
                            if (notOnOrAfter.before(new Date(System.currentTimeMillis()))) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("Audience")) {
                    if (!child.getFirstChild().getNodeValue().equals(resourceid)) {
                        throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                    }
                } else if (child.getNodeName().endsWith("AuthorizationDecisionStatement")) {
                    NamedNodeMap atts = child.getAttributes();
                    
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("Resource")) {
                            if (!atts.item(j).getTextContent().equals(resourceid)) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        } else if (atts.item(j).getNodeName().endsWith("Decision")) {
                            if (!atts.item(j).getTextContent().equals("Permit")) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("Signature")) {
                    validateSignature(child, false);
                }

                if (child.getNodeName() != null) {
                    if (child.getNodeName().endsWith("Evidence")) {
                        processAuthenticationToken(child, 0);
                    } else {
                        processAuthorizationToken(list.item(i), nextDebth);
                    }
                }
            }
        }
        if (authorizationIssuer == null && parsingDebth == 0) {
            throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
        }
    }

    /**
     * This method processes the authentication assertion that was in the header. It checks the
     * issuer, expiration dates and the authentication instant to see if this is still valid.
     * The signature element is checked by a separate method.
     */
    private void processAuthenticationToken(Node token, int parsingDebth)
            throws SOAPException, ServiceRequestException, GeneralSecurityException {

        int nextDebth = parsingDebth + 1;
        if (token.hasChildNodes()) {
            NodeList list = token.getChildNodes();
            Node child;
            
            for (int i = 0; i < list.getLength(); i++) {
                child = list.item(i);
                if (child.getNodeName() == null) {
                    /* Do nothing */
                } else if (child.getNodeName().endsWith("Assertion")) {
                    authenticationAssertion = child.cloneNode(true);
                    NamedNodeMap atts = child.getAttributes();
                    
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("Issuer")) {
                            authenticationIssuer = atts.item(j).getTextContent();
                        }
                    }
                } else if (child.getNodeName().endsWith("Conditions")) {
                    NamedNodeMap atts = child.getAttributes();
                    
                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("NotBefore")) {
                            Date notBefore = IsoDate.stringToDate(
                                    atts.item(j).getTextContent(), IsoDate.DATE_TIME);
                            if (notBefore.after(new Date(System.currentTimeMillis()))) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        } else if (atts.item(j).getNodeName().endsWith("NotOnOrAfter")) {
                            Date notOnOrAfter = IsoDate.stringToDate(
                                    atts.item(j).getTextContent(), IsoDate.DATE_TIME);
                            if (notOnOrAfter.before(new Date(System.currentTimeMillis()))) {
                                throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("AuthenticationStatement")) {
                    NamedNodeMap atts = child.getAttributes();

                    for (int j = 0; j < atts.getLength(); j++) {
                        if (atts.item(j).getNodeName().endsWith("AuthenticationInstant")) {
                            sessiontimeout = new Date(IsoDate.stringToDate(atts.item(j).getTextContent(), 
                                    IsoDate.DATE_TIME).getTime() + SESSION_TIME_TO_LIVE);
                            if (sessionTimedOut()) {
                                throw new ServiceRequestException(RegisterFault.TIMED_OUT);
                            }
                        }
                    }
                } else if (child.getNodeName().endsWith("Signature")) {
                    validateSignature(child, true);
                }
                processAuthenticationToken(list.item(i), nextDebth);
            }
        }
        if ((sessiontimeout == null || authenticationIssuer == null) && parsingDebth == 0) {
            throw new ServiceRequestException(RegisterFault.NOT_AUTHORIZED);
        }
    }

    /**
     * This method removes extra spaces, line breaks, wrong namespaces and the signature
     * element to make the assertion similar at all times when using it for signing and
     * signature verification. This method is used both for the authentication and
     * authorization assertion so the second parameter tells us how much should be removed.
     */
    private String prepareAssertion(String assertion, boolean removeOnlySecond) {
        StringBuffer buffer = new StringBuffer(assertion);

        /* Removes all multiple spaces */
        int doubleSpace = buffer.indexOf("  ");
        int somethingElse;
        while (doubleSpace > -1) {
            somethingElse = buffer.indexOf("<", doubleSpace);
            buffer.replace(doubleSpace, somethingElse, "");
            doubleSpace = buffer.indexOf("  ", doubleSpace);
        }

        /* Removes all \n */
        int linebreak = buffer.indexOf("\n");
        while (linebreak > -1) {
            buffer.replace(linebreak, linebreak + 1, "");
            linebreak = buffer.indexOf("\n", linebreak);
        }

        /* Removes all \r */
        linebreak = buffer.indexOf("\r");
        while (linebreak > -1) {
            buffer.replace(linebreak, linebreak + 1, "");
            linebreak = buffer.indexOf("\r", linebreak);
        }

        /* Replaces all ns1 with saml */
        int ns1index = buffer.indexOf("ns1:");
        while (ns1index > -1) {
            buffer.replace(ns1index, ns1index + 3, "saml");
            ns1index = buffer.indexOf("ns1:", ns1index);
        }
        ns1index = buffer.indexOf("ns1=");
        if (ns1index > -1) {
            buffer.replace(ns1index, ns1index + 3, "saml");
        }

        assertion = buffer.toString();

        /* Removes the signature element */
        int signatureStartIndex = assertion.indexOf("Signature");
        if (removeOnlySecond) {
            signatureStartIndex = assertion.indexOf("Signature>", signatureStartIndex + 1);
            signatureStartIndex = assertion.indexOf("Signature", signatureStartIndex + 1);
        }
        int tagStartIndex = assertion.indexOf("<", signatureStartIndex - 7);
        int lastTagStartIndex = 0;

        while (tagStartIndex < signatureStartIndex) {
            lastTagStartIndex = tagStartIndex;
            tagStartIndex = assertion.indexOf("<", lastTagStartIndex + 1);
        }
        tagStartIndex = lastTagStartIndex;
        int signatureEndIndex = assertion.indexOf("Signature>", signatureStartIndex + 1)
                + "Signature>".length();
        String result = assertion.substring(0, tagStartIndex)
                + assertion.substring(signatureEndIndex, assertion.length());

        /* Return the pruned string */
        return result;
    }

    /**
     * This method validates the signature and throws an Exception if there is something wrong,
     * to indicate that an invalid assertion and that we should not proceed with creating a
     * response.
     */
    private void validateSignature(Node signaturenode, boolean authentication)
            throws GeneralSecurityException {
        String clearText;

        /* Check whether this is the authentication or authorization assertion: */
        if (authentication) {
            clearText = prepareAssertion(authenticationAssertion.toString(), false);
        } else {
            clearText = prepareAssertion(authorizationAssertion.toString(), true);
        }

        NodeList children = signaturenode.getChildNodes();
        Node child;
        X509Certificate cert = null;
        String signatureString = null;

        for (int i = 0; i < children.getLength(); i++) {
            child = children.item(i);
            if (child.getNodeName().endsWith("SignatureValue")) {
                /* This node contains the actual signature */
                signatureString = child.getFirstChild().getNodeValue();
            } else if (child.getNodeName().endsWith("KeyInfo")) {
                /* This node contains information about the key used to create the signature */
                Node x509data = child.getFirstChild();
                Node x509certificate = x509data.getFirstChild();
                String certAsString = x509certificate.getFirstChild().getNodeValue();
                CertificateFactory certFactory = CertificateFactory.getInstance("X.509");
                ByteArrayInputStream certInput = new ByteArrayInputStream(Base64.decode(certAsString));
                cert = (X509Certificate) certFactory.generateCertificate(certInput);

                try {
                    certInput.close();
                } catch (IOException e) {
                }

                String certificateissuer = cert.getIssuerDN().getName();

                if (authentication && !authenticationIssuer.equals(certificateissuer)) {
                    throw new GeneralSecurityException("Authentication issuer: "
                            + authenticationIssuer + " Certificate issuer: "
                            + certificateissuer);
                } else if ((!authentication) && !authorizationIssuer.equals(certificateissuer)) {
                    throw new GeneralSecurityException("Authorization issuer: "
                            + authorizationIssuer + " Certificate issuer: "
                            + certificateissuer);
                }
            }
        }

        /* Signature verification: */
        Signature signer = Signature.getInstance("SHA1withRSA");
        signer.initVerify(cert);
        signer.update(clearText.getBytes());

        boolean ok = signer.verify(Base64.decode(signatureString));
        if (!ok) {
            throw new GeneralSecurityException(clearText);
        }
    }

    /**
     * This method runs through all the nodes in the request body to find the nodes we need,
     * which are the language and resourceID.
     */
    private void parseBody(Node node) throws SOAPException {
        if (node.hasChildNodes()) {
            NodeList list = node.getChildNodes();
            Node child;

            for (int i = 0; i < list.getLength(); i++) {
                child = list.item(i);
                if (child != null) {
                    if (child.getNodeName().endsWith("language")) {
                        language = child.getFirstChild().getNodeValue();
                    } else if (child.getNodeName().endsWith("ResourceID")) {
                        resourceid = child.getFirstChild().getNodeValue();
                        userid = resourceid.substring(resourceid.lastIndexOf("/") + 1);
                    } else {
                        parseBody(child);
                    }
                }
            }
        }
    }

    /**
     * @return String The language code or english if not set
     */
    public String getLanguage() {
        return (language != null ? language : RegisterService.LANGUAGECODE_GB);
    }

    /**
     * @return String The userid from the request
     */
    public String getUserID() {
        return userid;
    }

    /**
     * @return String The messageid of the request
     */
    public String getMessageID() {
        return messageid;
    }

    /**
     * @return Date The timestamp on the request
     */
    public Date getTimestamp() {
        return timestamp;
    }

    /**
     * @return boolean Returns true if the session has timed out and false if not
     */
    public boolean sessionTimedOut() {
        return sessiontimeout.before(new Date(System.currentTimeMillis()));
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (RegisterService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (RegisterService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
