/*
 * Classname: ServiceResponse
 * Version: 1.9
 * Date: 20.05.2006
 * Copyright: Copyright (c) 2006 Gunn Olaussen, Kirsti N. Torgersen. All Rights Reserved.
 * 
 * This software is the intellectual property of Gunn Olaussen and Kirsti N. Torgersen 
 * (henceforth called the developers). It cannot be exploited financially used for any 
 * purpose by others than the developers, except in accordance with the terms of the 
 * agreement between Kantega AS and the developers.
 */

package loanfund;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Date;
import javax.xml.soap.*;

import org.kobjects.isodate.IsoDate;

/**
 * The class ServiceResponse is used to create the reponse message. It needs to create the
 * header and the body. The body contains the service values. This also creates the tokens and
 * a signature of the assertion.
 * 
 * @author Gunn Olaussen & Kirsti N. Torgersen
 */
public class ServiceResponse {

    public static final int STATUS_INDEX = 6;
    public static final String GRANTED = "GRANTED";

    private static final String ns = "http://fedoralab1.kantega.no:8080/responses/mymobilesite-response.xsd";
    private static final String sb = "urn:liberty:sb:2003-08";

    private SOAPEnvelope env;
    private SOAPMessage message;
    private String inResponseTo;
    private String language;

    /* The service labels in English */
    private String[][] labelsGB = new String[][] {
            new String[] { "Total debt", "Sum" },
            new String[] { "Last instalment", "Date", "Sum" },
            new String[] { "Next instalment", "Date", "Sum" },
            new String[] { "Application status", "Date received", "Status", "Grant", "Loan", "Cause" } };

    /* The service labels in Norwegian */
    private String[][] labelsNB = new String[][] {
            new String[] { "Samlet gjeld", "Sum" },
            new String[] { "Forrige avdrag", "Dato", "Sum" },
            new String[] { "Neste avdrag", "Dato", "Sum" },
            new String[] { "Status p sknad", "Dato mottatt", "Status", "Stipend", "Ln", "rsak" } };

    /**
     * The constructor creates the message and initialises variables that are used in message
     * creation. It calls two other methods to create the header and body of the message.
     * 
     * @param inResponseTo String The messageid of the request this is a response to
     * @param language String The preferred language
     * @throws ServiceRequestException If an error occurs that indicate we should make a fault
     *         message instead
     */
    public ServiceResponse(String inResponseTo, String language) throws ServiceRequestException {
        this.language = language;
        this.inResponseTo = inResponseTo;
        try {
            message = MessageFactory.newInstance().createMessage();
            message.setProperty(SOAPMessage.CHARACTER_SET_ENCODING, "UTF-8");
            message.setProperty(SOAPMessage.WRITE_XML_DECLARATION, "true");

            env = message.getSOAPPart().getEnvelope();
            createHeader();
        } catch (SOAPException e) {
            if (RegisterService.DEBUG) Log.write(e);
            throw new ServiceRequestException(RegisterFault.INTERNAL_ERROR);
        }
    }

    /**
     * This method creates the header by adding all the required fields such as the messageID,
     * correlation and timestamp.
     */
    private void createHeader() throws SOAPException {
        SOAPHeader header = message.getSOAPHeader();
        SOAPHeaderElement correlation = header.addHeaderElement(env.createName("Correlation",
                "sb", sb));
        correlation.addAttribute(env.createName("messageID", "sb", sb), Randomizer.nextID());
        correlation.addAttribute(env.createName("refToMessageID", "sb", sb), inResponseTo);
        correlation.addAttribute(env.createName("timestamp", "sb", sb), IsoDate.dateToString(
                new Date(System.currentTimeMillis()), IsoDate.DATE_TIME));
        correlation.setMustUnderstand(true);
    }

    /**
     * When creating the body the method adds all the services in a ServiceList element which
     * holds the content of the response.
     */
    public void createBody(String[] values) throws SOAPException {
        String[][] labels;
        if (language.equals("nb")) {
            labels = labelsNB;
        } else {
            labels = labelsGB;
        }
        SOAPBody body = message.getSOAPBody();
        SOAPBodyElement serviceList = body.addBodyElement(env.createName("ServiceList", "ns0", ns));
        
        int x = 0;
        int tmpX = 0;
        boolean allNull = true;
        SOAPElement el;
        SOAPElement elchild;

        Name service = env.createName("Service", "ns0", ns);
        Name name = env.createName("name", "ns0", ns);
        Name valueElement = env.createName("ValueElement", "ns0", ns);
        Name label = env.createName("label", "ns0", ns);
        Name value = env.createName("value", "ns0", ns);

        String cause = getCauseText(values[STATUS_INDEX]);

        /* Adding the services: */
        for (int i = 0; i < labels.length; i++) {
            tmpX = x;
            allNull = true;
            for (int j = 1; j < labels[i].length && tmpX < values.length; j++) {
                if (values[tmpX] != null) allNull = false;
                tmpX++;
            }
            if (!allNull) {
                /* If the list contains services add them to the serviceList: */
                el = serviceList.addChildElement(service);
                el.addAttribute(name, labels[i][0]);
                for (int j = 1; j < labels[i].length && x < values.length; j++) {
                    if (values[x] != null && x != STATUS_INDEX) {
                        elchild = el.addChildElement(valueElement);
                        elchild.addAttribute(label, labels[i][j]);
                        elchild.addAttribute(value, values[x]);
                    } else if (values[x] != null && x == STATUS_INDEX) {
                        elchild = el.addChildElement(valueElement);
                        elchild.addAttribute(label, labels[i][j]);
                        elchild.addAttribute(value, getStatusText(values[x]));
                    }
                    x++;
                }
                if (x == values.length && cause != null) {
                    elchild = el.addChildElement(valueElement);
                    elchild.addAttribute(label,
                            labels[labels.length - 1][labels[labels.length - 1].length - 1]);
                    elchild.addAttribute(value, cause);
                }
            } else {
                x = tmpX;
            }
        }
    }

    /**
     * Returns the status text of the input code in the correct language
     */
    private String getStatusText(String statuscode) {
        if (statuscode == null) return null;
        if (statuscode.equals("GRANTED") && language.equals(RegisterService.LANGUAGECODE_GB)) return "Granted";
        if (statuscode.equals("GRANTED") && language.equals(RegisterService.LANGUAGECODE_NB)) return "Innvilget";
        if (statuscode.equals("REJECTED_NOTENROLLED") && language.equals(RegisterService.LANGUAGECODE_GB)) return "Rejected";
        if (statuscode.equals("REJECTED_NOTENROLLED") && language.equals(RegisterService.LANGUAGECODE_NB)) return "Avvist";
        if (statuscode.equals("REJECTED_MISSEDPAYMENT") && language.equals(RegisterService.LANGUAGECODE_GB)) return "Rejected";
        if (statuscode.equals("REJECTED_MISSEDPAYMENT") && language.equals(RegisterService.LANGUAGECODE_NB)) return "Avvist";
        if (statuscode.equals("PROCESSING") && language.equals(RegisterService.LANGUAGECODE_GB)) return "Processing";
        if (statuscode.equals("PROCESSING") && language.equals(RegisterService.LANGUAGECODE_NB)) return "Under behandling";
        return null;
    }

    /**
     * Returns the cause text of the input code in the correct language
     */
    private String getCauseText(String statuscode) {
        if (statuscode == null) return null;
        if (statuscode.equals("REJECTED_NOTENROLLED") && language.equals(RegisterService.LANGUAGECODE_GB))
            return "You were not enrolled with a qualified teaching facility at the time the application " +
                    "was processed.";
        if (statuscode.equals("REJECTED_NOTENROLLED") && language.equals(RegisterService.LANGUAGECODE_NB))
            return "Du var ikke registrert som student hos et godkjent lrested p det tidspunkt sknaden " +
                    "ble behandlet.";
        if (statuscode.equals("REJECTED_MISSEDPAYMENT") && language.equals(RegisterService.LANGUAGECODE_GB))
            return "You have unpayed installments.";
        if (statuscode.equals("REJECTED_MISSEDPAYMENT") && language.equals(RegisterService.LANGUAGECODE_NB))
            return "Du har ubetalte terminbelp.";
        return null;
    }

    /**
     * @return SOAPMessage The service response message
     */
    public SOAPMessage getMessage() {
        return message;
    }

    /**
     * Returns a string representation of this request.
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        try {
            message.writeTo(bytes);
            bytes.close();
            return bytes.toString();
        } catch (SOAPException e) {
            if (RegisterService.DEBUG) Log.write(e);
            return "SOAPException: " + e.getMessage();
        } catch (IOException e) {
            if (RegisterService.DEBUG) Log.write(e);
            return "IOException: " + e.getMessage();
        }
    }
}
